from numpy.typing import ArrayLike
from typing import Sequence
from typing import Tuple
from typing import Callable
from typing import Any
from numpy import ndarray
from typing import List
import matplotlib as mpl
import ubelt as ub
from _typeshed import Incomplete
from kwimage.structs import _generic

from kwimage._typing import SKImageGeometricTransform

__docstubs__: str


class Coords(_generic.Spatial, ub.NiceRepr):
    data: Incomplete
    meta: Incomplete

    def __init__(self,
                 data: Incomplete | None = ...,
                 meta: Incomplete | None = ...) -> None:
        ...

    def __array__(self):
        ...

    def __nice__(self):
        ...

    def __len__(self):
        ...

    @property
    def dtype(self):
        ...

    @property
    def dim(self):
        ...

    @property
    def shape(self):
        ...

    def copy(self):
        ...

    @classmethod
    def random(Coords,
               num: int = ...,
               dim: int = ...,
               rng: Incomplete | None = ...,
               meta: Incomplete | None = ...):
        ...

    def is_numpy(self) -> bool:
        ...

    def is_tensor(self) -> bool:
        ...

    def compress(self,
                 flags: ArrayLike,
                 axis: int = 0,
                 inplace: bool = False) -> Coords:
        ...

    def take(self,
             indices: ArrayLike,
             axis: int = 0,
             inplace: bool = False) -> Coords:
        ...

    def astype(self, dtype, inplace: bool = False) -> Coords:
        ...

    def round(self, decimals: int = 0, inplace: bool = False) -> Coords:
        ...

    def view(self, *shape) -> Coords:
        ...

    @classmethod
    def concatenate(cls, coords: Sequence[Coords], axis: int = 0) -> Coords:
        ...

    @property
    def device(self):
        ...

    def tensor(self, device=...) -> Coords:
        ...

    def numpy(self) -> Coords:
        ...

    def reorder_axes(self,
                     new_order: Tuple[int],
                     inplace: bool = False) -> Coords:
        ...

    def warp(self,
             transform: SKImageGeometricTransform | ArrayLike | Any | Callable,
             input_dims: Tuple | None = None,
             output_dims: Tuple | None = None,
             inplace: bool = False) -> Coords:
        ...

    def to_imgaug(self, input_dims) -> Any:
        ...

    @classmethod
    def from_imgaug(cls, kpoi):
        ...

    def scale(self,
              factor: float | Tuple[float, float],
              about: Tuple | None = None,
              output_dims: Tuple | None = None,
              inplace: bool = False) -> Coords:
        ...

    def translate(self,
                  offset: float | Tuple[float, float],
                  output_dims: Tuple | None = None,
                  inplace: bool = False) -> Coords:
        ...

    def rotate(self,
               theta: float,
               about: Tuple | None = None,
               output_dims: Tuple | None = None,
               inplace: bool = False) -> Coords:
        ...

    def fill(self,
             image,
             value,
             coord_axes: Tuple | None = None,
             interp: str = ...) -> ndarray:
        ...

    def soft_fill(self,
                  image,
                  coord_axes: Tuple | None = None,
                  radius: int = ...) -> ndarray:
        ...

    def draw_on(self,
                image: Incomplete | None = ...,
                fill_value: int = ...,
                coord_axes: Tuple = ...,
                interp: str = ...) -> ndarray:
        ...

    def draw(self,
             color: str = ...,
             ax: Incomplete | None = ...,
             alpha: Incomplete | None = ...,
             coord_axes: Tuple = ...,
             radius: int = ...,
             setlim: bool = False) -> List[mpl.collections.PatchCollection]:
        ...



"""
Contains functions that used to belong to im_cv2, but have been generalized to
accept different backends.
"""
from kwimage._backend_info import _default_backend


def warp_image(image, transform, dsize=None, antialias=False,
               interpolation='linear', border_mode=None, border_value=0,
               large_warp_dim=None, origin_convention='center',
               return_info=False, backend='auto'):
    """
    Applies an transformation to an image with optional antialiasing.

    Args:
        image (ndarray): the input image as a numpy array.
            Note: this is passed directly to cv2, so it is best to ensure that
            it is contiguous and using a dtype that cv2 can handle.

        transform (ndarray | dict | kwimage.Matrix): a coercable affine or
            projective matrix.  See :class:`kwimage.Affine` and
            :class:`kwimage.Projective` for details on what can be coerced.

        dsize (Tuple[int, int] | None | str):
            A integer width and height tuple of the resulting "canvas" image.
            If None, then the input image size is used.

            If specified as a string, dsize is computed based on the given
            heuristic.

            If 'positive' (or 'auto'), dsize is computed such that the positive
            coordinates of the warped image will fit in the new canvas. In this
            case, any pixel that maps to a negative coordinate will be clipped.
            This has the property that the input transformation is not
            modified.

            If 'content' (or 'max'), the transform is modified with an extra
            translation such that both the positive and negative coordinates of
            the warped image will fit in the new canvas.

        antialias (bool)
            if True determines if the transform is downsampling and applies
            antialiasing via gaussian a blur. Defaults to False

        interpolation (str | int):
            interpolation code or cv2 integer. Interpolation codes are linear,
            nearest, cubic, lancsoz, and area. Defaults to "linear".

        border_mode (str | int):
            Border code or cv2 integer. Border codes are constant (default)
            replicate, reflect, wrap, reflect101, and transparent.

        border_value (int | float | Iterable[int | float]):
            Used as the fill value if border_mode is constant. Otherwise this
            is ignored. Defaults to 0, but can also be defaulted to nan.
            if border_value is a scalar and there are multiple channels, the
            value is applied to all channels. More than 4 unique border values
            for individual channels will cause an error. See OpenCV #22283 for
            details.  In the future we may accept np.ma and return a masked
            array, but for now that is not implemented.

        large_warp_dim (int | None | str):
            If specified, perform the warp piecewise in chunks of the specified
            size. If "auto", it is set to the maximum "short" value in numpy.
            This works around a limitation of cv2.warpAffine, which must have
            image dimensions < SHRT_MAX (=32767 in version 4.5.3)

        origin_convention (str):
            Controls the interpretation of the underlying raster.
            Can be "center" (default opencv behavior), or "corner" (matches
            torchvision / detectron2 behavior).
            If "center", then center of the top left pixel is at (0, 0), and
            the top left corner is at (-0.5, -0.5).
            If "center", then center of the top left pixel is at (0.5, 0.5), and
            the top left corner is at (0, 0).
            Currently defaults to "center", but in the future we may change the
            default to "corner".  For more info see [WhereArePixels]_.

        return_info (bool):
            if True, returns information about the operation. In the case
            where dsize="content", this includes the modified transformation.

    Returns:
        ndarray | Tuple[ndarray, Dict]:
            the warped image, or if return info is True, the warped image and
            the info dictionary.

    SeeAlso:
        :func:`kwimage.warp_tensor`
        :func:`kwimage.warp_affine`
        :func:`kwimage.warp_projective`

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import kwimage
        >>> import ubelt as ub
        >>> image = kwimage.grab_test_image('paraview')
        >>> tf_homog = kwimage.Projective.random(rng=30342110) @ kwimage.Projective.coerce(uv=[0.001, 0.001])
        >>> tf_aff = kwimage.Affine.coerce(ub.udict(tf_homog.decompose()) - {'uv'})
        >>> tf_uv = kwimage.Projective.coerce(ub.udict(tf_homog.decompose()) & {'uv'})
        >>> warped1 = kwimage.warp_image(image, tf_homog, dsize='positive')
        >>> warped2 = kwimage.warp_image(image, tf_aff, dsize='positive')
        >>> warped3 = kwimage.warp_image(image, tf_uv, dsize='positive')
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> pnum_ = kwplot.PlotNums(nRows=2, nCols=2)
        >>> kwplot.imshow(warped1, pnum=pnum_(), title='projective warp')
        >>> kwplot.imshow(warped2, pnum=pnum_(), title='affine warp')
        >>> kwplot.imshow(warped3, pnum=pnum_(), title='projective part')
        >>> kwplot.show_if_requested()
    """
    import kwimage
    transform = kwimage.Projective.coerce(transform)
    kwargs = dict(dsize=dsize, antialias=antialias,
                  interpolation=interpolation, border_mode=border_mode,
                  border_value=border_value, large_warp_dim=large_warp_dim,
                  origin_convention=origin_convention,
                  return_info=return_info, backend=backend)
    if transform.is_affine():
        return kwimage.warp_affine(image, transform, **kwargs)
    else:
        return kwimage.warp_projective(image, transform, **kwargs)


def warp_projective(image, transform, dsize=None, antialias=False,
                    interpolation='linear', border_mode=None, border_value=0,
                    large_warp_dim=None, origin_convention='center',
                    return_info=False, backend='auto'):
    """
    Applies an projective transformation to an image with optional antialiasing.

    Args:
        image (ndarray): the input image as a numpy array.
            Note: this is passed directly to cv2, so it is best to ensure that
            it is contiguous and using a dtype that cv2 can handle.

        transform (ndarray | dict | kwimage.Projective): a coercable projective matrix.
            See :class:`kwimage.Projective` for details on what can be coerced.

        dsize (Tuple[int, int] | None | str):
            A integer width and height tuple of the resulting "canvas" image.
            If None, then the input image size is used.

            If specified as a string, dsize is computed based on the given
            heuristic.

            If 'positive' (or 'auto'), dsize is computed such that the positive
            coordinates of the warped image will fit in the new canvas. In this
            case, any pixel that maps to a negative coordinate will be clipped.
            This has the property that the input transformation is not
            modified.

            If 'content' (or 'max'), the transform is modified with an extra
            translation such that both the positive and negative coordinates of
            the warped image will fit in the new canvas.

        antialias (bool)
            if True determines if the transform is downsampling and applies
            antialiasing via gaussian a blur. Defaults to False

        interpolation (str | int):
            interpolation code or cv2 integer. Interpolation codes are linear,
            nearest, cubic, lancsoz, and area. Defaults to "linear".

        border_mode (str | int):
            Border code or cv2 integer. Border codes are constant (default)
            replicate, reflect, wrap, reflect101, and transparent.

        border_value (int | float | Iterable[int | float]):
            Used as the fill value if border_mode is constant. Otherwise this
            is ignored. Defaults to 0, but can also be defaulted to nan.
            if border_value is a scalar and there are multiple channels, the
            value is applied to all channels. More than 4 unique border values
            for individual channels will cause an error. See OpenCV #22283 for
            details.  In the future we may accept np.ma and return a masked
            array, but for now that is not implemented.

        large_warp_dim (int | None | str):
            If specified, perform the warp piecewise in chunks of the specified
            size. If "auto", it is set to the maximum "short" value in numpy.
            This works around a limitation of cv2.warpAffine, which must have
            image dimensions < SHRT_MAX (=32767 in version 4.5.3)

        origin_convention (str):
            Controls the interpretation of the underlying raster.
            Can be "center" (default opencv behavior), or "corner" (matches
            torchvision / detectron2 behavior).
            If "center", then center of the top left pixel is at (0, 0), and
            the top left corner is at (-0.5, -0.5).
            If "center", then center of the top left pixel is at (0.5, 0.5), and
            the top left corner is at (0, 0).
            Currently defaults to "center", but in the future we may change the
            default to "corner".  For more info see [WhereArePixels]_.

        return_info (bool):
            if True, returns information about the operation. In the case
            where dsize="content", this includes the modified transformation.

    Returns:
        ndarray | Tuple[ndarray, Dict]:
            the warped image, or if return info is True, the warped image and
            the info dictionary.
    """
    if backend == 'auto':
        backend = _default_backend()

    if backend == 'cv2':
        from kwimage.im_cv2 import _cv2_warp_projective
        return _cv2_warp_projective(
            image=image,
            transform=transform,
            dsize=dsize,
            antialias=antialias,
            interpolation=interpolation,
            border_mode=border_mode,
            border_value=border_value,
            large_warp_dim=large_warp_dim,
            origin_convention=origin_convention,
            return_info=return_info,
        )
    elif backend == 'itk':
        raise NotImplementedError(f'no kwimage backend={backend} for warp_projective')
    elif backend == 'torch':
        raise NotImplementedError(f'no kwimage backend={backend} for warp_projective')
    elif backend == 'skimage':
        raise NotImplementedError(f'no kwimage backend={backend} for warp_projective')
    else:
        raise KeyError(f'no kwimage backend={backend} for warp_projective')


def warp_affine(image, transform, dsize=None, antialias=False,
                interpolation='linear', border_mode=None, border_value=0,
                large_warp_dim=None, origin_convention='center',
                return_info=False, backend='auto'):
    """
    Applies an affine transformation to an image with optional antialiasing.

    Args:
        image (ndarray): the input image as a numpy array.
            Note: this is passed directly to cv2, so it is best to ensure that
            it is contiguous and using a dtype that cv2 can handle.

        transform (ndarray | dict | kwimage.Affine): a coercable affine matrix.
            See :class:`kwimage.Affine` for details on what can be coerced.

        dsize (Tuple[int, int] | None | str):
            A integer width and height tuple of the resulting "canvas" image.

            If None, then the input image size is used.

            If specified as a string, dsize is computed based on the given
            heuristic.

            If 'positive' (or 'auto'), dsize is computed such that the positive
            coordinates of the warped image will fit in the new canvas. In this
            case, any pixel that maps to a negative coordinate will be clipped.
            This has the property that the input transformation is not
            modified. NOTE: there are issues with this when the transformation
            includes rotation of reflections.

            If 'content' (or 'max'), the transform is modified with an extra
            translation such that both the positive and negative coordinates of
            the warped image will fit in the new canvas.

        antialias (bool)
            if True determines if the transform is downsampling and applies
            antialiasing via gaussian a blur. Defaults to False

        interpolation (str | int):
            interpolation code or cv2 integer. Interpolation codes are linear,
            nearest, cubic, lancsoz, and area. Defaults to "linear".

        border_mode (str | int):
            Border code or cv2 integer. Border codes are constant (default)
            replicate, reflect, wrap, reflect101, and transparent.

        border_value (int | float | Iterable[int | float] | str):
            Used as the fill value if border_mode is constant. Otherwise this
            is ignored. Defaults to 0, but can also be defaulted to nan.
            if border_value is a scalar and there are multiple channels, the
            value is applied to all channels. More than 4 unique border values
            for individual channels will cause an error. See OpenCV #22283 for
            details.  In the future we may accept np.ma and return a masked
            array, but for now that is not implemented.
            If a string, it indicates a border_mode that is not constant.
            In this case border mode should not be given.

        large_warp_dim (int | None | str):
            If specified, perform the warp piecewise in chunks of the specified
            size. If "auto", it is set to the maximum "short" value in numpy.
            This works around a limitation of cv2.warpAffine, which must have
            image dimensions < SHRT_MAX (=32767 in version 4.5.3)

        return_info (bool):
            if True, returns information about the operation. In the case
            where dsize="content", this includes the modified transformation.

        origin_convention (str):
            Controls the interpretation of the underlying raster.
            Can be "center" (default opencv behavior), or "corner" (matches
            torchvision / detectron2 behavior).
            If "center", then center of the top left pixel is at (0, 0), and
            the top left corner is at (-0.5, -0.5).
            If "center", then center of the top left pixel is at (0.5, 0.5), and
            the top left corner is at (0, 0).
            Currently defaults to "center", but in the future we may change the
            default to "corner".  For more info see [WhereArePixels]_.

    Returns:
        ndarray | Tuple[ndarray, Dict]:
            the warped image, or if return info is True, the warped image and
            the info dictionary.

    TODO:
        - [ ] When dsize='positive' but the transform contains an axis flip,
              the width / height of the box will become negative.  Should we
              adjust for this?

    References:
        .. [WhereArePixels] https://ppwwyyxx.com/blog/2021/Where-are-Pixels/

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import kwimage
        >>> from kwimage.transform import Affine
        >>> image = kwimage.grab_test_image('astro')
        >>> #image = kwimage.grab_test_image('checkerboard')
        >>> transform = Affine.random() @ Affine.scale(0.05)
        >>> transform = Affine.scale(0.02)
        >>> warped1 = kwimage.warp_affine(image, transform, dsize='positive', antialias=1, interpolation='nearest')
        >>> warped2 = kwimage.warp_affine(image, transform, dsize='positive', antialias=0)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> pnum_ = kwplot.PlotNums(nRows=1, nCols=2)
        >>> kwplot.imshow(warped1, pnum=pnum_(), title='antialias=True')
        >>> kwplot.imshow(warped2, pnum=pnum_(), title='antialias=False')
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import kwimage
        >>> from kwimage.transform import Affine
        >>> image = kwimage.grab_test_image('astro')
        >>> image = kwimage.grab_test_image('checkerboard')
        >>> transform = Affine.random() @ Affine.scale((.1, 1.2))
        >>> warped1 = kwimage.warp_affine(image, transform, dsize='positive', antialias=1)
        >>> warped2 = kwimage.warp_affine(image, transform, dsize='positive', antialias=0)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> pnum_ = kwplot.PlotNums(nRows=1, nCols=2)
        >>> kwplot.imshow(warped1, pnum=pnum_(), title='antialias=True')
        >>> kwplot.imshow(warped2, pnum=pnum_(), title='antialias=False')
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # Test the case where the input data is empty or the target canvas
        >>> # is empty, this should be handled like boundary effects
        >>> import kwimage
        >>> import numpy as np
        >>> image = np.random.rand(1, 1, 3)
        >>> transform = kwimage.Affine.random()
        >>> result = kwimage.warp_affine(image, transform, dsize=(0, 0))
        >>> assert result.shape == (0, 0, 3)
        >>> #
        >>> empty_image = np.random.rand(0, 1, 3)
        >>> result = kwimage.warp_affine(empty_image, transform, dsize=(10, 10))
        >>> assert result.shape == (10, 10, 3)
        >>> #
        >>> empty_image = np.random.rand(0, 1, 3)
        >>> result = kwimage.warp_affine(empty_image, transform, dsize=(10, 0))
        >>> assert result.shape == (0, 10, 3)

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # Demo difference between positive and content dsize
        >>> import kwimage
        >>> from kwimage.transform import Affine
        >>> image = kwimage.grab_test_image('astro', dsize=(512, 512))
        >>> transform = Affine.coerce(offset=(-100, -50), scale=2, theta=0.1)
        >>> # When warping other images or geometry along with this image
        >>> # it is important to account for the modified transform when
        >>> # setting dsize='content'. If dsize='positive', the transform
        >>> # will remain unchanged wrt other aligned images / geometries.
        >>> poly = kwimage.Boxes([[350, 5, 130, 290]], 'xywh').to_polygons()[0]
        >>> # Apply the warping to the images
        >>> warped_pos, info_pos = kwimage.warp_affine(image, transform, dsize='positive', return_info=True)
        >>> warped_con, info_con = kwimage.warp_affine(image, transform, dsize='content', return_info=True)
        >>> assert info_pos['dsize'] == (919, 1072)
        >>> assert info_con['dsize'] == (1122, 1122)
        >>> assert info_pos['transform'] == transform
        >>> # Demo the correct and incorrect way to apply transforms
        >>> poly_pos = poly.warp(transform)
        >>> poly_con = poly.warp(info_con['transform'])
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> # show original
        >>> kwplot.imshow(image, pnum=(1, 3, 1), title='original')
        >>> poly.draw(color='green', alpha=0.5, border=True)
        >>> # show positive warped
        >>> kwplot.imshow(warped_pos, pnum=(1, 3, 2), title='dsize=positive')
        >>> poly_pos.draw(color='purple', alpha=0.5, border=True)
        >>> # show content warped
        >>> ax = kwplot.imshow(warped_con, pnum=(1, 3, 3), title='dsize=content')[1]
        >>> poly_con.draw(color='dodgerblue', alpha=0.5, border=True)   # correct
        >>> poly_pos.draw(color='orangered', alpha=0.5, border=True)  # incorrect
        >>> cc = poly_con.to_shapely().centroid
        >>> cp = poly_pos.to_shapely().centroid
        >>> ax.text(cc.x, cc.y + 250, 'correctly transformed', color='dodgerblue',
        >>>         backgroundcolor=(0, 0, 0, 0.7), horizontalalignment='center')
        >>> ax.text(cp.x, cp.y - 250, 'incorrectly transformed', color='orangered',
        >>>         backgroundcolor=(0, 0, 0, 0.7), horizontalalignment='center')
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # Demo piecewise transform
        >>> import kwimage
        >>> from kwimage.transform import Affine
        >>> image = kwimage.grab_test_image('pm5644')
        >>> transform = Affine.coerce(offset=(-100, -50), scale=2, theta=0.1)
        >>> warped_piecewise, info = kwimage.warp_affine(image, transform, dsize='positive', return_info=True, large_warp_dim=32)
        >>> warped_normal, info = kwimage.warp_affine(image, transform, dsize='positive', return_info=True, large_warp_dim=None)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(image, pnum=(1, 3, 1), title='original')
        >>> kwplot.imshow(warped_normal, pnum=(1, 3, 2), title='normal warp')
        >>> kwplot.imshow(warped_piecewise, pnum=(1, 3, 3), title='piecewise warp')

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import kwimage
        >>> import numpy as np
        >>> # TODO: Explain why the bottom left is interpolated with 0's
        >>> # And not 2s, probably has to do with interpretation of pixels
        >>> # as points and not areas.
        >>> image = np.full((6, 6), fill_value=3, dtype=np.uint8)
        >>> transform = kwimage.Affine.eye()
        >>> transform = kwimage.Affine.coerce(offset=.5) @ transform
        >>> transform = kwimage.Affine.coerce(scale=2) @ transform
        >>> warped = kwimage.warp_affine(image, transform, dsize=(12, 12))

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # Demo how nans are handled
        >>> import kwimage
        >>> import numpy as np
        >>> image = kwimage.grab_test_image('pm5644')
        >>> image = kwimage.ensure_float01(image)
        >>> image[100:300, 400:700] = np.nan
        >>> transform = kwimage.Affine.coerce(scale=0.05, offset=10.5, theta=0.3, shearx=0.2)
        >>> warped1 = kwimage.warp_affine(image, transform, dsize='positive', antialias=1, interpolation='linear', border_value=0)
        >>> warped2 = kwimage.warp_affine(image, transform, dsize='positive', antialias=0, border_value=np.nan)
        >>> assert np.isnan(warped1).any()
        >>> assert np.isnan(warped2).any()
        >>> assert warped1[np.isnan(warped1).any(axis=2)].all()
        >>> assert warped2[np.isnan(warped2).any(axis=2)].all()
        >>> print('warped1.shape = {!r}'.format(warped1.shape))
        >>> print('warped2.shape = {!r}'.format(warped2.shape))
        >>> assert warped2.shape == warped1.shape
        >>> warped2[np.isnan(warped2).any(axis=2)]
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> pnum_ = kwplot.PlotNums(nRows=1, nCols=3)
        >>> image_canvas = kwimage.fill_nans_with_checkers(image)
        >>> warped1_canvas = kwimage.fill_nans_with_checkers(warped1)
        >>> warped2_canvas = kwimage.fill_nans_with_checkers(warped2)
        >>> kwplot.imshow(image_canvas, pnum=pnum_(), title='original')
        >>> kwplot.imshow(warped1_canvas, pnum=pnum_(), title='antialias=True, border=0')
        >>> kwplot.imshow(warped2_canvas, pnum=pnum_(), title='antialias=False, border=nan')
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # Demo how of how we also handle masked arrays
        >>> import kwimage
        >>> import numpy as np
        >>> _image = kwimage.grab_test_image('pm5644')
        >>> _image = kwimage.ensure_float01(_image)
        >>> _image[100:200, 400:700] = np.nan
        >>> mask = np.isnan(_image)
        >>> data = np.nan_to_num(_image)
        >>> image = np.ma.MaskedArray(data=data, mask=mask)
        >>> transform = kwimage.Affine.coerce(scale=0.05, offset=10.5, theta=0.3, shearx=0.2)
        >>> warped1 = kwimage.warp_affine(image, transform, dsize='positive', antialias=1, interpolation='linear')
        >>> assert isinstance(warped1, np.ma.MaskedArray)
        >>> warped2 = kwimage.warp_affine(image, transform, dsize='positive', antialias=0)
        >>> print('warped1.shape = {!r}'.format(warped1.shape))
        >>> print('warped2.shape = {!r}'.format(warped2.shape))
        >>> assert warped2.shape == warped1.shape
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> pnum_ = kwplot.PlotNums(nRows=1, nCols=2)
        >>> kwplot.imshow(warped1, pnum=pnum_(), title='antialias=True')
        >>> kwplot.imshow(warped2, pnum=pnum_(), title='antialias=False')
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import kwimage
        >>> import ubelt as ub
        >>> import numpy as np
        >>> #image = kwimage.checkerboard(dsize=(8, 8), num_squares=4, on_value='kitware_blue', off_value='kitware_green')
        >>> image = kwimage.checkerboard(dsize=(32, 32), num_squares=4, on_value='kitware_blue', off_value='kitware_green')
        >>> #image = kwimage.checkerboard(dsize=(4, 4), num_squares=4, on_value='kitware_blue', off_value='kitware_green')
        >>> grid = list(ub.named_product({
        >>>     'border_value': [
        >>>         0,
        >>>         np.nan,
        >>>         #'replicate'
        >>>    ],
        >>>     'interpolation': [
        >>>         'nearest',
        >>>         'linear',
        >>>         #'cubic',
        >>>         #'lanczos',
        >>>         #'area'
        >>>     ],
        >>>     'origin_convention': ['center', 'corner'],
        >>> }))
        >>> results = []
        >>> results += [('input', image)]
        >>> for kwargs in grid:
        >>>     warped_image = kwimage.warp_affine(image, {'scale': 0.5}, dsize='auto', **kwargs)
        >>>     title = ub.urepr(kwargs, compact=1, nl=1)
        >>>     results.append((title, warped_image))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> pnum_ = kwplot.PlotNums(nSubplots=len(results))
        >>> for title, canvas in results:
        >>>     canvas = kwimage.fill_nans_with_checkers(canvas).clip(0, 1)
        >>>     kwplot.imshow(canvas, pnum=pnum_(), origin_convention='corner', title=title, show_ticks=True)
        >>> fig = kwplot.plt.gcf()
        >>> fig.subplots_adjust(hspace=0.37, wspace=0.25)
        >>> fig.set_size_inches([18.11, 11.07])
        >>> kwplot.show_if_requested()
    """
    if backend == 'auto':
        backend = _default_backend()

    if backend == 'cv2':
        from kwimage.im_cv2 import _cv2_warp_affine
        return _cv2_warp_affine(
            image=image,
            transform=transform,
            dsize=dsize,
            antialias=antialias,
            interpolation=interpolation,
            border_mode=border_mode,
            border_value=border_value,
            large_warp_dim=large_warp_dim,
            origin_convention=origin_convention,
            return_info=return_info,
        )
    elif backend == 'itk':
        from kwimage.im_itk import _itk_warp_affine
        if large_warp_dim is not None:
            raise ValueError('itk backend does not support large_warp_dim')
        return _itk_warp_affine(
            image=image,
            transform=transform,
            dsize=dsize,
            antialias=antialias,
            interpolation=interpolation,
            border_mode=border_mode,
            border_value=border_value,
            origin_convention=origin_convention,
            return_info=return_info,
        )
    elif backend == 'torch':
        raise NotImplementedError(f'no kwimage backend={backend} for warp_affine')
    elif backend == 'skimage':
        raise NotImplementedError(f'no kwimage backend={backend} for warp_affine')
    else:
        raise KeyError(f'no kwimage backend={backend} for warp_affine')


def imresize(img, scale=None, dsize=None, max_dim=None, min_dim=None,
             interpolation=None, grow_interpolation=None, letterbox=False,
             return_info=False, antialias=False, border_value=0, backend='auto'):
    """
    Resize an image via a scale factor, final size, or size and aspect ratio.

    Wraps and generalizes cv2.resize, allows for specification of either a
    scale factor, a final size, or the final size for a particular dimension.

    Note:
        As described in [ResizeConfusion]_, this each entry in the image array
        as representing the center of a pixel. This is the pixels_are='area'
        approach, or align_corners=False in pytorch. It is equivalent to a
        shift and scale in warp_affine (which by default uses align corners).

    Note:
        The border mode cannot be specified here and seems to always be reflect
        in the underlying cv2 implementation.

    Args:
        img (ndarray): image to resize

        scale (float | Tuple[float, float]):
            Desired floating point scale factor. If a tuple, the dimension
            ordering is x,y.  Mutually exclusive with dsize, min_dim, max_dim.

        dsize (Tuple[int | None, int | None] | None):
            The desired width and height of the new image. If a dimension is
            None, then it is automatically computed to preserve aspect ratio.
            Mutually exclusive with scale, min_dim, max_dim.

        max_dim (int):
            New size of the maximum dimension, the other dimension is scaled to
            maintain aspect ratio. Mutually exclusive with scale, dsize,
            min_dim.

        min_dim (int):
            New size of the minimum dimension, the other dimension is scaled to
            maintain aspect ratio. Mutually exclusive with scale, dsize,
            max_dim.

        interpolation (str | int):
            The interpolation key or code (e.g. linear lanczos). By default
            "area" is used if the image is shrinking and "lanczos" is used if
            the image is growing. Note, if this is explicitly set, then it will
            be used regardless of if the image is growing or shrinking. Set
            ``grow_interpolation`` to change the default for an enlarging
            interpolation.

        grow_interpolation (str | int):
            The interpolation key or code to use when the image is being
            enlarged. Does nothing if "interpolation" is explicitly given. If
            "interpolation" is not specified "area" is used when shrinking.
            Defaults to "lanczos".

        letterbox (bool):
            If used in conjunction with dsize, then the image is scaled and
            translated to fit in the center of the new image while maintaining
            aspect ratio. Border padding is added if necessary.  Defaults to
            False.

        return_info (bool):
            if True returns information about the final transformation in a
            dictionary. If there is an offset, the scale is applied before the
            offset when transforming to the new resized space.  Defaults to
            False.

        antialias (bool):
            if True blurs to anti-alias before downsampling.
            Defaults to False.

        border_value (int | float | Iterable[int | float]):
            if letterbox is True, this is used as the constant fill value.

    Returns:
        ndarray | Tuple[ndarray, Dict] :
            the new image and optionally an info dictionary if
            `return_info=True`

    References:
        .. [ResizeConfusion] https://jricheimer.github.io/tensorflow/2019/02/11/resize-confusion/

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import kwimage
        >>> import numpy as np
        >>> # Test scale
        >>> img = np.zeros((16, 10, 3), dtype=np.uint8)
        >>> new_img, info = kwimage.imresize(img, scale=.85,
        >>>                                  interpolation='area',
        >>>                                  return_info=True)
        >>> print('info = {!r}'.format(info))
        >>> assert info['scale'].tolist() == [.8, 0.875]
        >>> # Test dsize without None
        >>> new_img, info = kwimage.imresize(img, dsize=(5, 12),
        >>>                                  interpolation='area',
        >>>                                  return_info=True)
        >>> print('info = {!r}'.format(info))
        >>> assert info['scale'].tolist() == [0.5 , 0.75]
        >>> # Test dsize with None
        >>> new_img, info = kwimage.imresize(img, dsize=(6, None),
        >>>                                  interpolation='area',
        >>>                                  return_info=True)
        >>> print('info = {!r}'.format(info))
        >>> assert info['scale'].tolist() == [0.6, 0.625]
        >>> # Test max_dim
        >>> new_img, info = kwimage.imresize(img, max_dim=6,
        >>>                                  interpolation='area',
        >>>                                  return_info=True)
        >>> print('info = {!r}'.format(info))
        >>> assert info['scale'].tolist() == [0.4  , 0.375]
        >>> # Test min_dim
        >>> new_img, info = kwimage.imresize(img, min_dim=6,
        >>>                                  interpolation='area',
        >>>                                  return_info=True)
        >>> print('info = {!r}'.format(info))
        >>> assert info['scale'].tolist() == [0.6  , 0.625]

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import kwimage
        >>> import numpy as np
        >>> # Test letterbox resize
        >>> img = np.ones((5, 10, 3), dtype=np.float32)
        >>> new_img, info = kwimage.imresize(img, dsize=(19, 19),
        >>>                                  letterbox=True,
        >>>                                  return_info=True)
        >>> print('info = {!r}'.format(info))
        >>> assert info['offset'].tolist() == [0, 4]
        >>> img = np.ones((10, 5, 3), dtype=np.float32)
        >>> new_img, info = kwimage.imresize(img, dsize=(19, 19),
        >>>                                  letterbox=True,
        >>>                                  return_info=True)
        >>> print('info = {!r}'.format(info))
        >>> assert info['offset'].tolist() == [4, 0]

        >>> import kwimage
        >>> import numpy as np
        >>> # Test letterbox resize
        >>> img = np.random.rand(100, 200)
        >>> new_img, info = kwimage.imresize(img, dsize=(300, 300), letterbox=True, return_info=True)

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # Check aliasing
        >>> import numpy as np
        >>> import kwimage
        >>> #img = kwimage.grab_test_image('checkerboard')
        >>> img = kwimage.grab_test_image('pm5644')
        >>> # test with nans
        >>> img = kwimage.ensure_float01(img)
        >>> img[100:200, 400:700] = np.nan
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> dsize = (14, 14)
        >>> dsize = (64, 64)
        >>> # When we set "grow_interpolation" for a "shrinking" resize it should
        >>> # still do the "area" interpolation to antialias the results. But if we
        >>> # use explicit interpolation it should alias.
        >>> pnum_ = kwplot.PlotNums(nSubplots=12, nCols=4)
        >>> kwplot.imshow(kwimage.imresize(img, dsize=dsize, antialias=True,  interpolation='area'), pnum=pnum_(), title='resize aa area')
        >>> kwplot.imshow(kwimage.imresize(img, dsize=dsize, antialias=True, interpolation='linear'), pnum=pnum_(), title='resize aa linear')
        >>> kwplot.imshow(kwimage.imresize(img, dsize=dsize, antialias=True, interpolation='nearest'), pnum=pnum_(), title='resize aa nearest')
        >>> kwplot.imshow(kwimage.imresize(img, dsize=dsize, antialias=True, interpolation='cubic'), pnum=pnum_(), title='resize aa cubic')

        >>> kwplot.imshow(kwimage.imresize(img, dsize=dsize, antialias=True, grow_interpolation='area'), pnum=pnum_(), title='resize aa grow area')
        >>> kwplot.imshow(kwimage.imresize(img, dsize=dsize, antialias=True, grow_interpolation='linear'), pnum=pnum_(), title='resize aa grow linear')
        >>> kwplot.imshow(kwimage.imresize(img, dsize=dsize, antialias=True, grow_interpolation='nearest'), pnum=pnum_(), title='resize aa grow nearest')
        >>> kwplot.imshow(kwimage.imresize(img, dsize=dsize, antialias=True, grow_interpolation='cubic'), pnum=pnum_(), title='resize aa grow cubic')

        >>> kwplot.imshow(kwimage.imresize(img, dsize=dsize, antialias=False, interpolation='area'), pnum=pnum_(), title='resize no-aa area')
        >>> kwplot.imshow(kwimage.imresize(img, dsize=dsize, antialias=False, interpolation='linear'), pnum=pnum_(), title='resize no-aa linear')
        >>> kwplot.imshow(kwimage.imresize(img, dsize=dsize, antialias=False, interpolation='nearest'), pnum=pnum_(), title='resize no-aa nearest')
        >>> kwplot.imshow(kwimage.imresize(img, dsize=dsize, antialias=False, interpolation='cubic'), pnum=pnum_(), title='resize no-aa cubic')

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # Test single pixel resize
        >>> import kwimage
        >>> import numpy as np
        >>> assert kwimage.imresize(np.random.rand(1, 1, 3), scale=3).shape == (3, 3, 3)
        >>> assert kwimage.imresize(np.random.rand(1, 1), scale=3).shape == (3, 3)

        # cv2.resize(np.random.rand(1, 1, 3), (3, 3))


    TODO:
        - [X] When interpolation is area and the number of channels > 4 cv2.resize will error but it is fine for linear interpolation
        - [ ] TODO: add padding options when letterbox=True

        - [ ] Allow for pre-clipping when letterbox=True
    """
    if backend == 'auto':
        backend = _default_backend()

    if backend == 'cv2':
        from kwimage.im_cv2 import _cv2_imresize
        return _cv2_imresize(
            img=img, scale=scale, dsize=dsize, max_dim=max_dim, min_dim=min_dim,
            interpolation=interpolation, grow_interpolation=grow_interpolation,
            letterbox=letterbox, return_info=return_info, antialias=antialias,
        )
    elif backend == 'skimage':
        return _skimage_resize(
            img=img, scale=scale, dsize=dsize, max_dim=max_dim, min_dim=min_dim,
            interpolation=interpolation, grow_interpolation=grow_interpolation,
            letterbox=letterbox, return_info=return_info, antialias=antialias,
        )
    elif backend == 'torch':
        raise NotImplementedError(backend)
    else:
        raise KeyError(backend)


def _skimage_resize(img, scale=None, dsize=None, max_dim=None, min_dim=None,
                    interpolation=None, grow_interpolation=None, letterbox=False,
                    return_info=False, antialias=False, border_value=0):
    """
    Example:
        >>> import numpy as np
        >>> from kwimage.im_transform import _skimage_resize
        >>> img = np.zeros((16, 10, 3), dtype=np.uint8)
        >>> new_img = _skimage_resize(img, dsize=(20, 21))
        >>> assert new_img.shape[0:2] == (21, 20)
    """
    from skimage.transform import resize
    not_implemented_params = [
        letterbox, min_dim, max_dim, scale, grow_interpolation,
        return_info, antialias
    ]
    for p in not_implemented_params:
        if p:
            raise NotImplementedError('skimage imresize params')

    order = _coerce_skimage_interpolation_order(interpolation)

    new_w, new_h = dsize
    output_shape = (new_h, new_w)
    new_img = resize(img, output_shape=output_shape, order=order)
    return new_img


def _coerce_skimage_interpolation_order(interpolation):
    """
    The order of interpolation. The order has to be in the range 0-5:
     - 0: Nearest-neighbor
     - 1: Bi-linear (default)
     - 2: Bi-quadratic
     - 3: Bi-cubic
     - 4: Bi-quartic
     - 5: Bi-quintic
    """
    if interpolation is None:
        interpolation = 'linear'   # TODO: make consistent with the cv2 default interpolation, which is currently lanczos

    if interpolation == 'nearest':
        order = 0
    elif interpolation == 'linear':
        order = 1
    elif interpolation == 'cubic':
        order = 3
    else:
        raise NotImplementedError(interpolation)
    return order



from typing import Any
from typing import Tuple
from typing import Callable
from numpy.typing import ArrayLike
import kwimage
from typing import List
from typing import Iterable
from numbers import Number
from numpy import ndarray
import shapely
from typing import Dict
import numpy as np
import matplotlib
import ubelt as ub
from _typeshed import Incomplete
from kwimage.structs import _generic

from kwimage._typing import SKImageGeometricTransform

__docstubs__: str


class _ShapelyMixin:

    def oriented_bounding_box(self):
        ...

    def buffer(self, *args, **kwargs):
        ...

    def simplify(self, tolerance, preserve_topology: bool = ...):
        ...

    @property
    def __geo_interface__(self):
        ...

    def union(self, other):
        ...

    def intersection(self, other):
        ...

    def difference(self, other):
        ...

    def symmetric_difference(self, other):
        ...

    def iooa(self, other):
        ...

    def iou(self, other):
        ...

    @property
    def area(self) -> float:
        ...

    @property
    def convex_hull(self):
        ...

    def is_invalid(self, explain: bool = False) -> bool | str:
        ...

    def fix(self):
        ...


class _PolyArrayBackend:

    def is_numpy(self):
        ...

    def is_tensor(self):
        ...

    def tensor(self, device=...):
        ...

    def numpy(self):
        ...


class _PolyWarpMixin:

    def to_imgaug(self, shape):
        ...

    def warp(self,
             transform: SKImageGeometricTransform | ArrayLike | Any | Callable,
             input_dims: Tuple | None = None,
             output_dims: Tuple | None = None,
             inplace: bool = False):
        ...

    def scale(self,
              factor: float | Tuple[float, float],
              about: Tuple | None = None,
              output_dims: Tuple | None = None,
              inplace: bool = False):
        ...

    def translate(self,
                  offset,
                  output_dims: Tuple | None = None,
                  inplace: bool = False):
        ...

    def rotate(self,
               theta: float,
               about: Tuple | None | str = None,
               output_dims: Tuple | None = None,
               inplace: bool = ...):
        ...

    def round(self, decimals: int = 0, inplace: bool = False) -> Polygon:
        ...

    def astype(self, dtype, inplace: bool = False) -> Polygon:
        ...

    def swap_axes(self, inplace: bool = False) -> Polygon:
        ...


class Polygon(_generic.Spatial, _PolyArrayBackend, _PolyWarpMixin,
              _ShapelyMixin, ub.NiceRepr):
    __datakeys__: Incomplete
    __metakeys__: Incomplete
    data: Incomplete
    meta: Incomplete

    def __init__(self,
                 data: Incomplete | None = ...,
                 meta: Incomplete | None = ...,
                 datakeys: Incomplete | None = ...,
                 metakeys: Incomplete | None = ...,
                 **kwargs) -> None:
        ...

    @property
    def exterior(self) -> kwimage.Coords:
        ...

    @property
    def interiors(self) -> List[kwimage.Coords]:
        ...

    def __nice__(self) -> str:
        ...

    @classmethod
    def circle(cls,
               xy: Iterable[Number] = ...,
               r: float | Number | Tuple[Number, Number] = 1.0,
               resolution: int = 64) -> Polygon:
        ...

    @classmethod
    def regular(cls, num, xy=..., r: int = ...):
        ...

    @classmethod
    def star(cls, xy=..., r: int = ...):
        ...

    @classmethod
    def random(cls,
               n: int = 6,
               n_holes: int = 0,
               convex: bool = True,
               tight: bool = False,
               rng: Incomplete | None = ...) -> Polygon:
        ...

    def to_mask(self,
                dims: Tuple | None = None,
                pixels_are: str = 'points') -> kwimage.Mask:
        ...

    def to_relative_mask(self, return_offset: bool = ...) -> kwimage.Mask:
        ...

    @classmethod
    def coerce(Polygon, data: object) -> kwimage.Polygon:
        ...

    @classmethod
    def from_shapely(
            Polygon,
            geom: shapely.geometry.polygon.Polygon) -> kwimage.Polygon:
        ...

    @classmethod
    def from_wkt(Polygon, data: str) -> kwimage.Polygon:
        ...

    @classmethod
    def from_geojson(Polygon, data_geojson: dict) -> Polygon:
        ...

    def to_shapely(self,
                   fix: bool = False) -> shapely.geometry.polygon.Polygon:
        ...

    def to_geojson(self) -> Dict[str, object]:
        ...

    def to_wkt(self) -> str:
        ...

    @classmethod
    def from_coco(cls,
                  data: List[Number] | Dict,
                  dims: None | Tuple[int, ...] = None) -> Polygon:
        ...

    def to_coco(self, style: str = 'orig') -> List | Dict:
        ...

    def to_multi_polygon(self) -> MultiPolygon:
        ...

    def to_boxes(self) -> kwimage.Boxes:
        ...

    @property
    def centroid(self) -> Tuple[Number, Number]:
        ...

    def to_box(self) -> kwimage.Box:
        ...

    def bounding_box(self) -> kwimage.Boxes:
        ...

    def box(self) -> kwimage.Box:
        ...

    def bounding_box_polygon(self) -> kwimage.Polygon:
        ...

    def copy(self) -> Polygon:
        ...

    def clip(self, x_min, y_min, x_max, y_max, inplace: bool = ...) -> Polygon:
        ...

    def fill(self,
             image: ndarray,
             value: int | Tuple[int] = 1,
             pixels_are: str = 'points',
             assert_inplace: bool = False) -> ndarray:
        ...

    def draw_on(self,
                image: ndarray,
                color: str | tuple = 'blue',
                fill: bool = True,
                border: bool = False,
                alpha: float = 1.0,
                edgecolor: str | tuple | None = None,
                facecolor: str | tuple | None = None,
                copy: bool = False) -> np.ndarray:
        ...

    def draw(
            self,
            color: str | Tuple = 'blue',
            ax: Incomplete | None = ...,
            alpha: float = 1.0,
            radius: int = ...,
            setlim: bool | str = False,
            border: bool | None = None,
            linewidth: bool | None = None,
            edgecolor: None | Any = None,
            facecolor: None | Any = None,
            fill: bool = True,
            vertex: float = False,
            vertexcolor: Any | None = None
    ) -> matplotlib.patches.PathPatch | None:
        ...

    def interpolate(self, other, alpha):
        ...

    def morph(self, other: kwimage.Polygon,
              alpha: float | List[float]) -> Polygon | List[Polygon]:
        ...


class MultiPolygon(_generic.ObjectList, _ShapelyMixin):

    @classmethod
    def random(self,
               n: int = ...,
               n_holes: int = ...,
               rng: Incomplete | None = ...,
               tight: bool = ...) -> MultiPolygon:
        ...

    def fill(self,
             image: ndarray,
             value: int | Tuple[int, ...] = 1,
             pixels_are: str = ...,
             assert_inplace: bool = ...) -> ndarray:
        ...

    def to_multi_polygon(self) -> MultiPolygon:
        ...

    def to_boxes(self) -> kwimage.Boxes:
        ...

    def to_box(self) -> kwimage.Box:
        ...

    def bounding_box(self) -> kwimage.Boxes:
        ...

    def box(self) -> kwimage.Box:
        ...

    def to_mask(self,
                dims: Incomplete | None = ...,
                pixels_are: str = ...) -> kwimage.Mask:
        ...

    def to_relative_mask(self, return_offset: bool = ...) -> kwimage.Mask:
        ...

    @classmethod
    def coerce(cls,
               data,
               dims: Incomplete | None = ...) -> None | MultiPolygon:
        ...

    def to_shapely(self, fix: bool = False) -> shapely.geometry.MultiPolygon:
        ...

    @classmethod
    def from_shapely(
        MultiPolygon,
        geom: shapely.geometry.MultiPolygon | shapely.geometry.Polygon
    ) -> MultiPolygon:
        ...

    @classmethod
    def from_geojson(MultiPolygon, data_geojson: Dict) -> MultiPolygon:
        ...

    def to_geojson(self) -> Dict:
        ...

    @classmethod
    def from_coco(cls,
                  data: List[List[Number] | Dict],
                  dims: None | Tuple[int, ...] = None) -> MultiPolygon:
        ...

    def to_coco(self, style: str = 'orig'):
        ...

    def swap_axes(self, inplace: bool = False) -> MultiPolygon:
        ...

    def draw_on(self, image, **kwargs):
        ...


class PolygonList(_generic.ObjectList):

    def to_mask_list(self,
                     dims: Incomplete | None = ...,
                     pixels_are: str = ...) -> kwimage.MaskList:
        ...

    def to_polygon_list(self) -> PolygonList:
        ...

    def to_segmentation_list(self) -> kwimage.SegmentationList:
        ...

    def swap_axes(self, inplace: bool = ...) -> PolygonList:
        ...

    def to_geojson(self, as_collection: bool = False) -> List[Dict] | Dict:
        ...

    def fill(self,
             image: ndarray,
             value: int | Tuple[int, ...] = 1,
             pixels_are: str = ...,
             assert_inplace: bool = ...) -> ndarray:
        ...

    def draw_on(self, *args, **kw):
        ...

    def unary_union(self):
        ...



"""
Logic pertaining to run-length encodings. Can encode an ndarray as a RLE or
decode an RLE into an ndarray.

SeeAlso:
    kwimage.structs.mask - stores binary segmentation masks, using RLEs as a
        backend representation. Also contains cython logic for handling
        the coco-rle format.
"""
import numpy as np


def encode_run_length(img, binary=False, order='C'):
    """
    Construct the run length encoding (RLE) of an image.

    Args:
        img (ndarray): 2D image

        binary (bool):
            If true, assume that the input image only contains 0's and 1's.
            Set to True for compatibility with COCO (which does not support
            multi-value RLE encodings).

        order (str):
            Order of the encoding.  Either 'C' for row major or 'F' for
            column-major. Defaults to 'C'.

    Returns:
        Dict[str, object]: encoding: dictionary items are:

            counts (ndarray): the run length encoding

            shape (Tuple): the original image shape.
                This should be in standard shape row-major (e.g.  h/w) order

            binary (bool):
                if True, the counts are assumed to encode only 0's and 1's,
                otherwise the counts encoding specifies any numeric values.

            order (str):
                Encoding order, either 'C' for row major or 'F' for
                column-major. Defaults to 'C'.

    SeeAlso:
        :class:`kwimage.Mask` -
            cython-backed data structure to handle coco-style RLEs

    Example:
        >>> import ubelt as ub
        >>> lines = ub.codeblock(
        >>>     '''
        >>>     ..........
        >>>     ......111.
        >>>     ..2...111.
        >>>     .222..111.
        >>>     22222.....
        >>>     .222......
        >>>     ..2.......
        >>>     ''').replace('.', '0').splitlines()
        >>> img = np.array([list(map(int, line)) for line in lines])
        >>> encoding = encode_run_length(img)
        >>> target = np.array([0,16,1,3,0,3,2,1,0,3,1,3,0,2,2,3,0,2,1,3,0,1,2,5,0,6,2,3,0,8,2,1,0,7])
        >>> assert np.all(target == encoding['counts'])

    Example:
        >>> binary = True
        >>> img = np.array([[1, 0, 1, 1, 1, 0, 0, 1, 0]])
        >>> encoding = encode_run_length(img, binary=True)
        >>> assert encoding['counts'].tolist() == [0, 1, 1, 3, 2, 1, 1]

    Example:
        >>> # Test empty case
        >>> from kwimage.im_runlen import *  # NOQA
        >>> binary = True
        >>> img = np.zeros((0, 0), dtype=np.uint8)
        >>> encoding = encode_run_length(img, binary=True)
        >>> assert encoding['counts'].tolist() == []
        >>> recon = decode_run_length(**encoding)
        >>> assert np.all(recon == img)

    Example:
        >>> # Test small full cases
        >>> for d in [0, 1, 2, 3]:
        >>>     img = np.zeros((d, d), dtype=np.uint8)
        >>>     encoding = encode_run_length(img, binary=True)
        >>>     recon = decode_run_length(**encoding)
        >>>     assert np.all(recon == img)
        >>>     img = np.ones((d, d), dtype=np.uint8)
        >>>     encoding = encode_run_length(img, binary=True)
        >>>     recon = decode_run_length(**encoding)
        >>>     assert np.all(recon == img)
    """
    flat = img.ravel(order=order)

    if flat.size == 0:
        values = np.empty((0,), dtype=int)
        lengths = np.empty((0,), dtype=int)
    else:
        diff_idxs = np.flatnonzero(np.abs(np.diff(flat)) > 0)
        pos = np.hstack([[0], diff_idxs + 1])
        lengths = np.hstack([np.diff(pos), [len(flat) - pos[-1]]])
        values = flat[pos]

    if binary:
        # Assume there are only zeros and ones
        if not set(np.unique(values)).issubset({0, 1}):
            raise ValueError('more than 0 and 1')

        if len(values) and values[0] != 0:
            # the binary RLE always starts with zero
            counts = np.hstack([[0], lengths])
        else:
            counts = lengths
    else:
        counts = np.hstack([values[:, None], lengths[:, None]]).ravel()

    encoding = {
        'shape': img.shape,
        'counts': counts,
        'binary': binary,
        'order': order,
    }
    return encoding


def decode_run_length(counts, shape, binary=False, dtype=np.uint8, order='C'):
    """
    Decode run length encoding back into an image.

    Args:
        counts (ndarray): the run-length encoding

        shape (Tuple[int, int]): the height / width of the mask

        binary (bool): if the RLE is binary or non-binary.
            Set to True for compatibility with COCO.

        dtype (type):
            data type for decoded image.  Defaults to np.uint8.

        order (str):
            Order of the encoding.  Either 'C' for row major or 'F' for
            column-major. Defaults to 'C'.

    Returns:
        ndarray: the reconstructed image

    Example:
        >>> from kwimage.im_runlen import *  # NOQA
        >>> img = np.array([[1, 0, 1, 1, 1, 0, 0, 1, 0]])
        >>> encoded = encode_run_length(img, binary=True)
        >>> recon = decode_run_length(**encoded)
        >>> assert np.all(recon == img)

        >>> import ubelt as ub
        >>> lines = ub.codeblock(
        >>>     '''
        >>>     ..........
        >>>     ......111.
        >>>     ..2...111.
        >>>     .222..111.
        >>>     22222.....
        >>>     .222......
        >>>     ..2.......
        >>>     ''').replace('.', '0').splitlines()
        >>> img = np.array([list(map(int, line)) for line in lines])
        >>> encoded = encode_run_length(img)
        >>> recon = decode_run_length(**encoded)
        >>> assert np.all(recon == img)
    """
    recon = np.zeros(shape, dtype=dtype, order=order)
    flat = recon.ravel(order)
    if binary:
        value = 0
        start = 0
        for num in counts:
            stop = start + num
            flat[start:stop] = value
            # print('value, start, start = {}, {}, {}'.format(value, start, stop))
            start = stop
            value = 1 - value
    else:
        start = 0
        for value, num in zip(counts[::2], counts[1::2]):
            stop = start + num
            flat[start:stop] = value
            start = stop
    return recon


def rle_translate(rle, offset, output_shape=None):
    """
    Translates a run-length encoded image in RLE-space.

    Args:
        rle (dict):
            an enconding dict returned by :func:`kwimage.encode_run_length`

        offset (Tuple[int, int]):
            x, y integer offsets.

        output_shape (Tuple[int, int]):
            h,w of transformed mask.
            If unspecified the input rle shape is used.

    SeeAlso:
        # ITK has some RLE code that looks like it can perform translations
        https://github.com/KitwareMedical/ITKRLEImage/blob/master/include/itkRLERegionOfInterestImageFilter.h

    Doctest:
        >>> # test that translate works on all zero images
        >>> img = np.zeros((7, 8), dtype=np.uint8)
        >>> rle = encode_run_length(img, binary=True, order='F')
        >>> new_rle = rle_translate(rle, (1, 2), (6, 9))
        >>> assert np.all(new_rle['counts'] == [54])

    Example:
        >>> from kwimage.im_runlen import *  # NOQA
        >>> img = np.array([
        >>>     [1, 1, 1, 1],
        >>>     [0, 1, 0, 0],
        >>>     [0, 1, 0, 1],
        >>>     [1, 1, 1, 1],], dtype=np.uint8)
        >>> rle = encode_run_length(img, binary=True, order='C')
        >>> offset = (1, -1)
        >>> output_shape = (3, 5)
        >>> new_rle = rle_translate(rle, offset, output_shape)
        >>> decoded = decode_run_length(**new_rle)
        >>> print(decoded)
        [[0 0 1 0 0]
         [0 0 1 0 1]
         [0 1 1 1 1]]

    Example

        >>> from kwimage.im_runlen import *  # NOQA
        >>> img = np.array([
        >>>     [0, 0, 0],
        >>>     [0, 1, 0],
        >>>     [0, 0, 0]], dtype=np.uint8)
        >>> rle = encode_run_length(img, binary=True, order='C')
        >>> new_rle = rle_translate(rle, (1, 0))
        >>> decoded = decode_run_length(**new_rle)
        >>> print(decoded)
        [[0 0 0]
         [0 0 1]
         [0 0 0]]
        >>> new_rle = rle_translate(rle, (0, 1))
        >>> decoded = decode_run_length(**new_rle)
        >>> print(decoded)
        [[0 0 0]
         [0 0 0]
         [0 1 0]]
    """
    if set(rle.keys()) == {'size', 'counts'}:
        # Handle coco rle's
        rle = rle.copy()
        rle['shape'] = rle['size']
        rle['order'] = 'F'
        rle['binary'] = True

    if not rle['binary']:
        raise NotImplementedError(
            'only binary rle translation is implemented')

    # Careful of residuals
    orig_offset = np.array(offset)
    offset = np.round(orig_offset).astype(int)
    residual = orig_offset - offset.astype(orig_offset.dtype)

    if not np.all(np.abs(residual) < 1e-6):
        import warnings
        warnings.warn('translating by rle, but offset is non-integer')

    # These are the flat indices where the value changes:
    #  * even locs are stop-indices for zeros and start indices for ones
    #  * odd locs are stop-indices for ones and start indices for zeros
    try:
        indices = rle['counts'].cumsum()
    except AttributeError:
        indices = np.array(rle['counts']).cumsum()

    if len(indices) % 2 == 1:
        indices = indices[:-1]

    # Transform indices to be start-stop inclusive indices for ones
    indices[1::2] -= 1

    # Find yx points where the binary mask changes value
    old_shape = np.array(rle['shape'])
    if output_shape is None:
        output_shape = old_shape
    new_shape = np.array(output_shape)
    rc_offset = np.array(offset[::-1])

    pts = np.unravel_index(indices, old_shape, order=rle['order'])
    major_axis = 1 if rle['order'] == 'F' else 0
    minor_axis = 1 - major_axis

    major_idxs = pts[major_axis]

    # Find locations in the major axis points where a non-zero count
    # crosses into the next minor dimension.
    pair_major_index = major_idxs.reshape(-1, 2)
    num_major_crossings = pair_major_index.T[1] - pair_major_index.T[0]
    flat_cross_idxs = np.where(num_major_crossings > 0)[0] * 2

    # Insert breaks to runs in locations that cross the major-axis.
    # This will force all runs to exist only within a single major-axis.
    # IE: Ensure points don't span multiple scanlines
    scanline_pts = [x.tolist() for x in pts]
    for idx in flat_cross_idxs[::-1]:
        prev_pt = [x[idx] for x in scanline_pts]
        next_pt = [x[idx + 1] for x in scanline_pts]
        for break_d in reversed(range(prev_pt[major_axis], next_pt[major_axis])):
            # Insert a breakpoint over every major axis crossing
            if major_axis == 1:
                new_stop = [old_shape[0] - 1, break_d]
                new_start = [0, break_d + 1]
            elif major_axis == 0:
                new_stop = [break_d, old_shape[1] - 1]
                new_start = [break_d + 1, 0]
            else:
                raise AssertionError(major_axis)
            scanline_pts[0].insert(idx + 1, new_start[0])
            scanline_pts[1].insert(idx + 1, new_start[1])
            scanline_pts[0].insert(idx + 1, new_stop[0])
            scanline_pts[1].insert(idx + 1, new_stop[1])

    # Now that new start-stop locations have been added,
    # translate the points that indices where non-zero data should go.
    new_pts = np.array(scanline_pts, dtype=int) + rc_offset[:, None]

    # <handle_out_of_bounds>
    # Note: all of the following logic relies on the fact that each run can
    # only span one major axis. This condition is true because we inserted
    # breakpoints whenever a run spanned more than one major axis.

    new_major_dim = new_shape[major_axis]
    new_minor_dim = new_shape[minor_axis]

    # Only keep points where the major axis is in bounds
    _new_major_pts = new_pts[major_axis]
    is_major_ib = ((_new_major_pts >= 0) &
                   (_new_major_pts < new_major_dim))
    # assert np.all(is_major_ib[0::2] == is_major_ib[1::2]), (
    #     'all pairs should be both in-bounds or both out-of-bounds')
    new_pts = new_pts.T[is_major_ib].T
    new_pts = np.ascontiguousarray(new_pts)

    # Now remove any points where the minor axis is OOB in the same direction.
    # (i.e. remove pairs of points that are both left-oob or both right-oob,
    # but dont remove pairs where only one is left-oob or right-oob, because
    # these still create structure in our new image.)
    _new_minor_pts = new_pts[minor_axis]
    is_left_oob = (_new_minor_pts < 0)
    is_right_oob = (_new_minor_pts >= new_minor_dim)
    is_pair_left_oob = (is_left_oob[0::2] & is_left_oob[1::2])
    is_pair_right_oob = (is_right_oob[0::2] & is_right_oob[1::2])
    is_pair_removable = (is_pair_left_oob | is_pair_right_oob)
    new_pts_pairs = new_pts.T.reshape(-1, 2, 2)
    new_pts = new_pts_pairs[~is_pair_removable].reshape(-1, 2).T
    new_pts = np.ascontiguousarray(new_pts)

    # Finally, all new points are strictly within the existing major dims and
    # we have removed any point pair where both pairs were oob in the same
    # direction, we can simply clip any regions along the minor axis that go
    # out of bounds.
    _new_minor_pts = new_pts[minor_axis]
    _new_minor_pts.clip(0, new_minor_dim - 1, out=_new_minor_pts)
    # </handle_out_of_bounds>

    # Now we have translated flat-indices in the new canvas shape
    new_indices = np.ravel_multi_index(new_pts, new_shape, order=rle['order'])
    new_indices[1::2] += 1

    count_dtype = int  # use in to eventually support non-binary RLE
    new_indices = new_indices.astype(count_dtype)

    total = int(np.prod(new_shape))
    if len(new_indices) == 0:
        trailing_counts = np.array([total], dtype=count_dtype)
        leading_counts = np.array([], dtype=count_dtype)
    else:
        leading_counts = np.array([new_indices[0]], dtype=count_dtype)
        trailing_counts = np.array([total - new_indices[-1]], dtype=count_dtype)

    body_counts = np.diff(new_indices)
    new_counts = np.hstack([leading_counts, body_counts, trailing_counts])

    new_rle = {
        'shape': tuple(new_shape.tolist()),
        'order': rle['order'],
        'counts': new_counts,
        'binary': rle['binary'],
    }
    return new_rle


def _rle_bytes_to_array(s, impl='auto'):
    """
    Uncompresses a coco-bytes RLE into an array representation.

    Args:
        s (bytes): compressed coco bytes rle
        impl (str): which implementation to use (defaults to cython is possible)

    CommandLine:
        xdoctest -m ~/code/kwimage/kwimage/im_runlen.py _rle_bytes_to_array

    Benchmark:
        >>> import ubelt as ub
        >>> from kwimage.im_runlen import _rle_bytes_to_array
        >>> s = b';?1B10O30O4'
        >>> ti = ub.Timerit(1000, bestof=50, verbose=2)
        >>> # --- time python impl ---
        >>> for timer in ti.reset('python'):
        >>>     with timer:
        >>>         _rle_bytes_to_array(s, impl='python')
        >>> # --- time cython impl ---
        >>> # xdoctest: +REQUIRES(--mask)
        >>> for timer in ti.reset('cython'):
        >>>     with timer:
        >>>         _rle_bytes_to_array(s, impl='cython')
    """
    # verbatim inefficient impl.
    # It would be nice if this (un/)compression algo could get a better
    # description.
    from kwimage.structs.mask import _backends
    key, cython_mask = _backends.get_backend(['kwimage'])

    if impl == 'auto':
        if cython_mask is None:
            impl = 'python'
        else:
            impl = 'cython'
    if impl == 'python':
        import numpy as np
        cnts = np.empty(len(s), dtype=np.int64)
        p = 0
        m = 0
        for m in range(len(s)):
            if p >= len(s):
                break
            x = 0
            k = 0
            more = 1
            while more:
                # c = s[p] - 48
                c = s[p] - 48
                x |= (c & 0x1f) << 5 * k
                more = c & 0x20
                p += 1
                k += 1
                if more == 0 and (c & 0x10):
                    x |= (-1 << 5 * k)
            if m > 2:
                x += cnts[m - 2]
            cnts[m] = x
        cnts = cnts[:m]
        return cnts
    elif impl == 'cython':
        return cython_mask._rle_bytes_to_array(s)


def _rle_array_to_bytes(counts, impl='auto'):
    """
    Compresses an array RLE into a coco-bytes RLE.

    Args:
        counts (ndarray): uncompressed array rle
        impl (str): which implementation to use (defaults to cython is possible)

    Example:
        >>> # xdoctest: +REQUIRES(--mask)
        >>> from kwimage.im_runlen import _rle_array_to_bytes
        >>> from kwimage.im_runlen import _rle_bytes_to_array
        >>> arr_counts = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
        >>> str_counts = _rle_array_to_bytes(arr_counts)
        >>> arr_counts2 = _rle_bytes_to_array(str_counts)
        >>> assert np.all(arr_counts2 == arr_counts)

    Benchmark:
        >>> # xdoctest: +REQUIRES(--mask)
        >>> import ubelt as ub
        >>> from kwimage.im_runlen import _rle_array_to_bytes
        >>> from kwimage.im_runlen import _rle_bytes_to_array
        >>> counts = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
        >>> ti = ub.Timerit(1000, bestof=50, verbose=2)
        >>> # --- time python impl ---
        >>> #for timer in ti.reset('python'):
        >>> #    with timer:
        >>> #        _rle_array_to_bytes(s, impl='python')
        >>> # --- time cython impl ---
        >>> for timer in ti.reset('cython'):
        >>>     with timer:
        >>>         _rle_array_to_bytes(s, impl='cython')
    """
    # verbatim inefficient impl.
    # It would be nice if this (un/)compression algo could get a better
    # description.
    from kwimage.structs.mask import _backends
    key, cython_mask = _backends.get_backend(['kwimage'])
    if impl == 'auto':
        if cython_mask is None:
            impl = 'python'
        else:
            impl = 'cython'

    if impl == 'python':
        raise NotImplementedError('pure python rle is not available')
    elif impl == 'cython':
        counts = counts.astype(np.uint32)
        counts_str = cython_mask._rle_array_to_bytes(counts)
        return counts_str
    else:
        raise KeyError(impl)


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m kwimage.im_runlen
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
Data structures to represent and manipulate 2D Points
"""
import numpy as np
import ubelt as ub
import kwarray
import numbers
import warnings
from kwimage.structs import _generic


__docstubs__ = """
from kwimage._typing import SKImageGeometricTransform
"""


class _PointsWarpMixin:

    def _warp_imgaug(self, augmenter, input_dims, inplace=False):
        """
        Warps by applying an augmenter from the imgaug library

        Args:
            augmenter (imgaug.augmenters.Augmenter):
            input_dims (Tuple): h/w of the input image
            inplace (bool): if True, modifies data inplace

        Example:
            >>> # xdoctest: +REQUIRES(module:imgaug)
            >>> from kwimage.structs.points import *  # NOQA
            >>> import imgaug
            >>> input_dims = (10, 10)
            >>> self = Points.random(10).scale(input_dims)
            >>> augmenter = imgaug.augmenters.Fliplr(p=1)
            >>> new = self._warp_imgaug(augmenter, input_dims)

            >>> self = Points(xy=(np.random.rand(10, 2) * 10).astype(int))
            >>> augmenter = imgaug.augmenters.Fliplr(p=1)
            >>> new = self._warp_imgaug(augmenter, input_dims)

            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> plt = kwplot.autoplt()
            >>> kwplot.figure(fnum=1, doclf=True)
            >>> ax = plt.gca()
            >>> ax.set_xlim(0, 10)
            >>> ax.set_ylim(0, 10)
            >>> self.draw(color='red', alpha=.4, radius=0.1)
            >>> new.draw(color='blue', alpha=.4, radius=0.1)
        """
        new = self if inplace else self.__class__(self.data.copy(), self.meta)
        new.data['xy'] = new.data['xy']._warp_imgaug(augmenter, input_dims,
                                                     inplace=inplace)
        if 'tf_data_to_img' in self.meta:
            # warping via imgaug invalidates the tf_data_to_img transform
            self.meta = self.meta.copy()
            self.meta.pop('tf_data_to_img')
        return new

    def to_imgaug(self, input_dims):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:imgaug)
            >>> from kwimage.structs.points import *  # NOQA
            >>> pts = Points.random(10)
            >>> input_dims = (10, 10)
            >>> kpoi = pts.to_imgaug(input_dims)
        """
        return self.data['xy'].to_imgaug(input_dims)

    @classmethod
    def from_imgaug(cls, kpoi):
        import kwimage
        data = kwimage.Coords.from_imgaug(kpoi)
        self = cls(data)
        return self

    @property
    def dtype(self):
        try:
            return self.data.dtype
        except Exception:
            print('kwimage.mask: no dtype for ' + str(type(self.data)))
            raise

    def warp(self, transform, input_dims=None, output_dims=None, inplace=False):
        """
        Generalized coordinate transform.

        Args:
            transform (ArrayLike | Callable | kwimage.Affine | SKImageGeometricTransform | Augmenter):
                scikit-image tranform, a 3x3 transformation matrix,
                an imgaug Augmenter, or generic callable which transforms
                an NxD ndarray.

            input_dims (Tuple): shape of the image these objects correspond to
                (only needed / used when transform is an imgaug augmenter)

            output_dims (Tuple): unused, only exists for compatibility

            inplace (bool): if True, modifies data inplace

        Example:
            >>> from kwimage.structs.points import *  # NOQA
            >>> import skimage
            >>> self = Points.random(10, rng=0)
            >>> transform = skimage.transform.AffineTransform(scale=(2, 2))
            >>> new = self.warp(transform)
            >>> assert np.all(new.xy == self.scale(2).xy)

        Doctest:
            >>> self = Points.random(10, rng=0)
            >>> assert np.all(self.warp(np.eye(3)).xy == self.xy)
            >>> assert np.all(self.warp(np.eye(2)).xy == self.xy)
        """
        import kwimage
        import skimage
        from kwimage._typing import SKImageGeometricTransform
        new = self if inplace else self.__class__(self.data.copy(), self.meta)
        if transform is None:
            return new

        if not isinstance(transform, (np.ndarray,
                                      SKImageGeometricTransform,
                                      kwimage.Affine)):
            try:
                import imgaug
            except ImportError:
                pass
                # warnings.warn('imgaug is not installed')
                # raise TypeError(type(transform))
            else:
                if isinstance(transform, imgaug.augmenters.Augmenter):
                    return new._warp_imgaug(transform, input_dims, inplace=True)
            # else:
            #     raise TypeError(type(transform))
        new.data['xy'] = new.data['xy'].warp(transform, input_dims,
                                             output_dims, inplace)
        if 'tf_data_to_img' in new.meta:
            # if we are maintaining a transform to img space, we need to update it
            new.meta = new.meta.copy()
            tf = transform
            if isinstance(tf, np.ndarray):
                tf = skimage.transform.AffineTransform(matrix=transform)
            elif callable(tf):
                raise NotImplementedError(
                    'callables cant transform linear data_to_img yet')
            inv_tf = skimage.transform.AffineTransform(matrix=tf._inv_matrix)
            # new.meta['tf_data_to_img'] = new.meta['tf_data_to_img'] + inv_tf
            new.meta['tf_data_to_img'] = inv_tf + new.meta['tf_data_to_img']
        return new

    def scale(self, factor, output_dims=None, inplace=False):
        """
        Scale a points by a factor

        Args:
            factor (float | Tuple[float, float]):
                scale factor as either a scalar or a (sf_x, sf_y) tuple.
            output_dims (Tuple): unused in non-raster spatial structures

        Example:
            >>> from kwimage.structs.points import *  # NOQA
            >>> self = Points.random(10, rng=0)
            >>> new = self.scale(10)
            >>> assert new.xy.max() <= 10
        """
        import skimage
        new = self if inplace else self.__class__(self.data.copy(), self.meta)
        new.data['xy'] = new.data['xy'].scale(factor, output_dims=output_dims,
                                              inplace=inplace)
        if 'tf_data_to_img' in new.meta:
            # if we are maintaining a transform to img space, we need to update it
            new.meta = new.meta.copy()
            tf = skimage.transform.AffineTransform(scale=factor)
            inv_tf = skimage.transform.AffineTransform(matrix=tf._inv_matrix)
            new.meta['tf_data_to_img'] = (inv_tf + new.meta['tf_data_to_img'])
        return new

    def translate(self, offset, output_dims=None, inplace=False):
        """
        Shift the points

        Args:
            factor (float | Tuple[float, float]):
                transation amount as either a scalar or a (t_x, t_y) tuple.
            output_dims (Tuple): unused in non-raster spatial structures

        Example:
            >>> from kwimage.structs.points import *  # NOQA
            >>> self = Points.random(10, rng=0)
            >>> new = self.translate(10)
            >>> assert new.xy.min() >= 10
            >>> assert new.xy.max() <= 11
        """
        import skimage
        new = self if inplace else self.__class__(self.data.copy(), self.meta)
        new.data['xy'] = new.data['xy'].translate(offset, output_dims, inplace)
        if 'tf_data_to_img' in new.meta:
            # if we are maintaining a transform to img space, we need to update it
            new.meta = new.meta.copy()
            tf = skimage.transform.AffineTransform(translation=offset)
            inv_tf = skimage.transform.AffineTransform(matrix=tf._inv_matrix)
            new.meta['tf_data_to_img'] = (inv_tf + new.meta['tf_data_to_img'])
        return new


class Points(_generic.Spatial, _PointsWarpMixin):
    """
    Stores multiple keypoints for a single object.

    This stores both the geometry and the class metadata if available

    Ignore:
        meta = {
         "names" = ['head', 'nose', 'tail'],
         "skeleton" = [(0, 1), (0, 2)],
        }

    Example:
        >>> from kwimage.structs.points import *  # NOQA
        >>> xy = np.random.rand(10, 2)
        >>> pts = Points(xy=xy)
        >>> print('pts = {!r}'.format(pts))
    """
    # __slots__ = ('data', 'meta',)

    # Pre-registered keys for the data dictionary
    __datakeys__ = ['xy', 'class_idxs', 'visible']
    # Pre-registered keys for the meta dictionary
    __metakeys__ = ['classes']

    def __init__(self, data=None, meta=None, datakeys=None, metakeys=None,
                 **kwargs):
        if kwargs:
            if data or meta:
                raise ValueError('Cannot specify kwargs AND data/meta dicts')
            _datakeys = self.__datakeys__
            _metakeys = self.__metakeys__
            # Allow the user to specify custom data and meta keys
            if datakeys is not None:
                _datakeys = _datakeys + list(datakeys)
            if metakeys is not None:
                _metakeys = _metakeys + list(metakeys)
            # Perform input checks whenever kwargs is given
            data = {key: kwargs.pop(key) for key in _datakeys if key in kwargs}
            meta = {key: kwargs.pop(key) for key in _metakeys if key in kwargs}
            if kwargs:
                raise ValueError(
                    'Unknown kwargs: {}'.format(sorted(kwargs.keys())))

            if 'xy' in data:
                if _generic.isinstance_arraytypes(data['xy']):
                    import kwimage
                    data['xy'] = kwimage.Coords(data['xy'])

        elif isinstance(data, self.__class__):
            # Avoid runtime checks and assume the user is doing the right thing
            # if data and meta are explicitly specified
            meta = data.meta
            data = data.data
        if meta is None:
            meta = {}
        self.data = data
        self.meta = meta

    def __nice__(self):
        data_repr = repr(self.xy)
        if '\n' in data_repr:
            data_repr = ub.indent('\n' + data_repr.lstrip('\n'), '    ')
        return 'xy={}'.format(data_repr)

    __repr__ = ub.NiceRepr.__str__

    def __len__(self):
        return len(self.data['xy'])

    @property
    def shape(self):
        return self.data['xy'].shape

    @property
    def xy(self):
        return self.data['xy'].data

    @classmethod
    def random(Points, num=1, classes=None, rng=None):
        """
        Makes random points; typically for testing purposes

        Example:
            >>> import kwimage
            >>> self = kwimage.Points.random(classes=[1, 2, 3])
            >>> self.data
            >>> print('self.data = {!r}'.format(self.data))
        """
        rng = kwarray.ensure_rng(rng)
        if ub.iterable(num):
            shape = tuple(num) + (2,)
        else:
            shape = (num, 2)
        self = Points(xy=rng.rand(*shape))
        self.data['visible'] = np.full(len(self), fill_value=2)
        if classes is not None:
            class_idxs = (rng.rand(len(self)) * len(classes)).astype(int)
            self.data['class_idxs'] = class_idxs
            self.meta['classes'] = classes
        return self

    def is_numpy(self):
        return self.data['xy'].is_numpy()

    def is_tensor(self):
        return self.data['xy'].is_tensor()

    @ub.memoize_property
    def _impl(self):
        return self.data['xy']._impl

    def tensor(self, device=ub.NoParam):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> from kwimage.structs.points import *  # NOQA
            >>> self = Points.random(10)
            >>> self.tensor()
        """
        impl = self._impl
        newdata = {k: v.tensor(device) if hasattr(v, 'tensor')
                   else impl.tensor(v, device)
                   for k, v in self.data.items()}
        new = self.__class__(newdata, self.meta)
        return new

    def round(self, inplace=False):
        """
        Rounds data to the nearest integer

        Args:
            inplace (bool): if True, modifies this object

        Example:
            >>> import kwimage
            >>> self = kwimage.Points.random(3).scale(10)
            >>> self.round()
        """
        new = self if inplace else self.__class__(self.data, self.meta)
        new.data['xy'] = self.data['xy'].round()
        return new

    def numpy(self):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> from kwimage.structs.points import *  # NOQA
            >>> self = Points.random(10)
            >>> self.tensor().numpy().tensor().numpy()
        """
        impl = self._impl
        newdata = {k: v.numpy() if hasattr(v, 'numpy') else impl.numpy(v)
                   for k, v in self.data.items()}
        new = self.__class__(newdata, self.meta)
        return new

    def draw_on(self, image=None, color='white', radius=None, copy=False):
        """

        Args:
            image (ndarray): image to draw points on.

            color (str | Any | List[Any]):
                one color for all boxes or a list of colors for each box
                Can be any type accepted by kwimage.Color.coerce.
                Extended types: str | ColorLike | List[ColorLike]

            radius (None | int):
                if an integer, an circle is drawn at each xy point with this
                radius.
                if None, attempts to fill a single point with subpixel accuracy,
                which generally means 4 pixels will be given some weight.
                Note: color can only be a single value for all points in this
                case.

            copy (bool): if True, force a copy of the image, otherwise
                try to draw inplace (may not work depending on dtype).

        CommandLine:
            xdoctest -m ~/code/kwimage/kwimage/structs/points.py Points.draw_on --show

        Example:
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> # xdoctest: +REQUIRES(module:cv2)
            >>> from kwimage.structs.points import *  # NOQA
            >>> s = 128
            >>> image = np.zeros((s, s))
            >>> self = Points.random(10).scale(s)
            >>> image = self.draw_on(image)
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.figure(fnum=1, doclf=True)
            >>> kwplot.autompl()
            >>> kwplot.imshow(image)
            >>> self.draw(radius=3, alpha=.5)
            >>> kwplot.show_if_requested()

        Example:
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> from kwimage.structs.points import *  # NOQA
            >>> s = 128
            >>> image = np.zeros((s, s))
            >>> self = Points.random(10).scale(s)
            >>> image = self.draw_on(image, radius=3, color='distinct')
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.figure(fnum=1, doclf=True)
            >>> kwplot.autompl()
            >>> kwplot.imshow(image)
            >>> #self.draw(radius=3, alpha=.5, color='classes')
            >>> kwplot.show_if_requested()

        Example:
            >>> # xdoctest: +REQUIRES(module:cv2)
            >>> import kwimage
            >>> s = 32
            >>> self = kwimage.Points.random(10).scale(s)
            >>> color = 'kitware_green'
            >>> # Test drawing on all channel + dtype combinations
            >>> im3 = np.zeros((s, s, 3), dtype=np.float32)
            >>> im_chans = {
            >>>     'im3': im3,
            >>>     'im1': kwimage.convert_colorspace(im3, 'rgb', 'gray'),
            >>>     'im4': kwimage.convert_colorspace(im3, 'rgb', 'rgba'),
            >>> }
            >>> inputs = {}
            >>> for k, im in im_chans.items():
            >>>     inputs[k + '_01'] = (kwimage.ensure_float01(im.copy()), {'radius': None})
            >>>     inputs[k + '_255'] = (kwimage.ensure_uint255(im.copy()), {'radius': None})
            >>> outputs = {}
            >>> for k, v in inputs.items():
            >>>     im, kw = v
            >>>     outputs[k] = self.draw_on(im, color=color, **kw)
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.figure(fnum=2, doclf=True)
            >>> plt = kwplot.autoplt()
            >>> pnum_ = kwplot.PlotNums(nRows=2, nSubplots=len(inputs))
            >>> for k in inputs.keys():
            >>>     kwplot.imshow(outputs[k], fnum=2, pnum=pnum_(), title=k)
            >>> plt.gcf().suptitle('Test draw points on channel + dtype combos')
            >>> kwplot.show_if_requested()

        Example:
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> from kwimage.structs.points import *  # NOQA
            >>> self = Points.random(10).scale(32)
            >>> image = self.draw_on(radius=3, color='distinct')
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> kwplot.imshow(image)
            >>> kwplot.show_if_requested()

        Example:
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> # Test cases where single and multiple colors are given
            >>> # with radius=None and radius=scalar
            >>> from kwimage.structs.points import *  # NOQA
            >>> import kwimage
            >>> self = kwimage.Points.random(10).scale(32)
            >>> image1 = self.draw_on(radius=2, color='blue')
            >>> image2 = self.draw_on(radius=None, color='blue')
            >>> image3 = self.draw_on(radius=2, color='distinct')
            >>> image4 = self.draw_on(radius=None, color='distinct')
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> canvas = kwimage.stack_images_grid(
            >>>     [image1, image2, image3, image4],
            >>>     pad=3, bg_value=(1, 1, 1))
            >>> kwplot.autompl()
            >>> kwplot.imshow(canvas)
            >>> kwplot.show_if_requested()
        """
        import kwimage
        if image is None:
            maxx, maxy = self.xy.max(axis=0)
            maxx = int(np.ceil(maxx) + 1)
            maxy = int(np.ceil(maxy) + 1)
            image = np.zeros((maxy, maxx, 3), dtype=np.float32)
        elif isinstance(image, tuple):
            # I forgot what the standard is that we use here...
            maxy, maxx = image
            image = np.zeros((maxx, maxy, 3), dtype=np.float32)

        if 0 in image.shape[0:2]:
            # Cannot draw on this image without width and height, return it as is
            return image

        dtype_fixer = _generic._consistent_dtype_fixer(image)

        single_color = False

        if color == 'distinct':
            colors = [kwimage.Color(c) for c in kwimage.Color.distinct(len(self))]
        elif color == 'classes':
            # TODO: read colors from categories if they exist
            class_idxs = self.data['class_idxs']
            _keys, _vals = kwarray.group_indices(class_idxs)
            cls_colors = kwimage.Color.distinct(len(self.meta['classes']))
            colors = list(ub.take(cls_colors, class_idxs))
            colors = [kwimage.Color(c) for c in colors]
        else:
            num = len(self)
            if isinstance(color, list) and not isinstance(color, numbers.Number):
                # Passed list of color for each point
                colors = [kwimage.Color(c) for c in color]
            else:
                # Passed a single color
                single_color = True
                colors = [kwimage.Color(color)] * num

        if radius is None:
            image = kwimage.atleast_3channels(image)
            image = kwimage.ensure_float01(image, copy=copy)
            # value = kwimage.Color(color).as01()
            if single_color:
                color_value = np.array(colors[0].forimage(image))
                image = self.data['xy'].fill(
                    image, color_value, coord_axes=[1, 0], interp='bilinear')
            else:
                # Need to loop when thare are multiple colors
                color_values = [np.array(kwimage.Color(c).forimage(image))
                                for c in colors]
                xy_pts = self.data['xy'].data.reshape(-1, 2)
                for xy, color_ in zip(xy_pts, color_values):
                    image = kwimage.subpixel_setvalue(
                        image, xy[None, :], color_, coord_axes=[1, 0],
                        interp='bilinear')
        else:
            import cv2
            image = kwimage.atleast_3channels(image, copy=copy)
            # note: ellipse has a different return type (UMat) and does not
            # work inplace if the input is not contiguous.
            image = np.ascontiguousarray(image)

            xy_pts = self.data['xy'].data.reshape(-1, 2)
            color_values = [kwimage.Color(c).forimage(image) for c in colors]

            for xy, color_ in zip(xy_pts, color_values):
                # center = tuple(map(int, xy.tolist()))
                center = tuple(xy.tolist())
                axes = (radius / 2, radius / 2)
                center = tuple(map(int, center))
                axes = tuple(map(int, axes))
                # print('center = {!r}'.format(center))
                # print('axes = {!r}'.format(axes))

                cv2.ellipse(image, center, axes, angle=0.0, startAngle=0.0,
                            endAngle=360.0, color=color_, thickness=-1)

        image = dtype_fixer(image, copy=False)
        return image

    def draw(self, color='blue', ax=None, alpha=None, radius=1, setlim=False, **kwargs):
        """
        TODO: can use kwplot.draw_points

        Example:
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> from kwimage.structs.points import *  # NOQA
            >>> pts = Points.random(10)
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.figure(doclf=1)
            >>> pts.draw(radius=0.01)
            >>> from kwimage.structs.points import *  # NOQA
            >>> self = Points.random(10, classes=['a', 'b', 'c'])
            >>> self.draw(radius=0.01, color='classes')
        """
        import kwimage
        import matplotlib as mpl
        from matplotlib import pyplot as plt
        if ax is None:
            ax = plt.gca()
        xy = self.data['xy'].data.reshape(-1, 2)
        # kwplot.draw_points(color=color, class_idxs)

        # More grouped patches == more efficient runtime
        if alpha is None:
            alpha = [1.0] * len(xy)
        elif not ub.iterable(alpha):
            alpha = [alpha] * len(xy)

        if color == 'distinct':
            colors = kwimage.Color.distinct(len(alpha))
        elif color == 'classes':
            # TODO: read colors from categories if they exist
            try:
                class_idxs = self.data['class_idxs']
                cls_colors = kwimage.Color.distinct(len(self.meta['classes']))
            except KeyError:
                raise Exception('cannot draw class colors without class_idxs and classes')
            _keys, _vals = kwarray.group_indices(class_idxs)
            colors = list(ub.take(cls_colors, class_idxs))
        else:
            colors = [color] * len(alpha)

        ptcolors = [kwimage.Color(c, alpha=a).as01('rgba')
                    for c, a in zip(colors, alpha)]
        color_groups = ub.group_items(range(len(ptcolors)), ptcolors)

        circlekw = {
            'radius': radius,
            'fill': True,
            'ec': None,
        }
        if 'fc' in kwargs:
            warnings.warning(
                'Warning: specifying fc to Points.draw overrides '
                'the color argument. Use color instead')
        circlekw.update(kwargs)
        fc = circlekw.pop('fc', None)  # hack

        collections = []
        for pcolor, idxs in color_groups.items():

            # hack for fc
            if fc is not None:
                pcolor = fc

            print(f'circlekw={circlekw}')
            print(f'pcolor={pcolor}')
            patches = [
                mpl.patches.Circle((x, y), fc=pcolor, **circlekw)
                for x, y in xy[idxs]
            ]
            col = mpl.collections.PatchCollection(patches, match_original=True)
            collections.append(col)
            ax.add_collection(col)

        if setlim:
            xmin = xy[:, 0].min()
            xmax = xy[:, 0].max()
            ymin = xy[:, 1].min()
            ymax = xy[:, 1].max()
            _generic._setlim(xmin, ymin, xmax, ymax, setlim=setlim, ax=ax)

        return collections

    def compress(self, flags, axis=0, inplace=False):
        """
        Filters items based on a boolean criterion

        Example:
            >>> from kwimage.structs.points import *  # NOQA
            >>> self = Points.random(4)
            >>> flags = [1, 0, 1, 1]
            >>> other = self.compress(flags)
            >>> assert len(self) == 4
            >>> assert len(other) == 3

            >>> # xdoctest: +REQUIRES(module:torch)
            >>> other = self.tensor().compress(flags)
            >>> assert len(other) == 3
        """
        new = self if inplace else self.__class__(self.data.copy(), self.meta)
        for k, v in self.data.items():
            try:
                new.data[k] = _generic._safe_compress(v, flags, axis=axis)
            except Exception:
                print('FAILED TO COMPRESS k={!r}, v={!r}'.format(k, v))
                raise
        return new

    def take(self, indices, axis=0, inplace=False):
        """
        Takes a subset of items at specific indices

        Example:
            >>> from kwimage.structs.points import *  # NOQA
            >>> self = Points.random(4)
            >>> indices = [1, 3]
            >>> other = self.take(indices)
            >>> assert len(self) == 4
            >>> assert len(other) == 2

            >>> # xdoctest: +REQUIRES(module:torch)
            >>> other = self.tensor().take(indices)
            >>> assert len(other) == 2
        """
        new = self if inplace else self.__class__(self.data.copy(), self.meta)
        for k, v in self.data.items():
            new.data[k] = _generic._safe_take(v, indices, axis=axis)
        return new

    @classmethod
    def concatenate(cls, points, axis=0):
        if len(points) == 0:
            raise ValueError('need at least one box to concatenate')
        if axis != 0:
            raise ValueError('can only concatenate along axis=0')
        import kwimage
        first = points[0]
        datas = [p.data['xy'] for p in points]
        newxy = kwimage.Coords.concatenate(datas)
        new = cls({'xy': newxy}, first.meta)
        return new

    def to_coco(self, style='orig'):
        """
        Converts to an mscoco-like representation

        Note:
            items that are usually id-references to other objects may need to
            be rectified.

        Args:
            style (str): either orig, new, new-id, new-name, or new-v2

        Returns:
            Dict: mscoco-like representation

        Example:
            >>> from kwimage.structs.points import *  # NOQA
            >>> self = Points.random(4, classes=['a', 'b'])
            >>> orig = self._to_coco(style='orig')
            >>> print('orig = {!r}'.format(orig))
            >>> new_name = self._to_coco(style='new-name')
            >>> print('new_name = {}'.format(ub.urepr(new_name, nl=-1)))
            >>> # xdoctest: +REQUIRES(module:kwcoco)
            >>> import kwcoco
            >>> self.meta['classes'] = kwcoco.CategoryTree.coerce(self.meta['classes'])
            >>> new_id = self._to_coco(style='new-id')
            >>> print('new_id = {}'.format(ub.urepr(new_id, nl=-1)))
            >>> # Test new-v2 style
            >>> new_v2 = self._to_coco(style='new-v2')
            >>> print('new_v2 = {}'.format(ub.urepr(new_v2, nl=-1)))
        """
        if self.xy.size == 0:
            return []
        if len(self.xy.shape) == 2:
            return self._to_coco(style=style)
        else:
            raise NotImplementedError('dim > 2, dense case todo')

    def _to_coco(self, style='orig'):
        """
        See to_coco
        """
        if style == 'orig':
            visible = self.data.get('visible', None)
            assert len(self.xy.shape) == 2
            if visible is None:
                visible = np.full((len(self), 1), fill_value=2)
            else:
                visible = visible.reshape(-1, 1)

            # TODO: ensure these are in the right order for the classes
            flat_pts = np.hstack([self.xy, visible]).reshape(-1)
            return flat_pts.tolist()
        elif style.startswith('new-v2'):
            new_kpts_v2 = {}
            x, y = self.data['xy'].data.T.tolist()
            new_kpts_v2['x'] = x
            new_kpts_v2['y'] = y
            class_idxs = self.data.get('class_idxs', None)
            if class_idxs is not None:
                assert 'classes' in self.meta
                classes = self.meta['classes']
                try:
                    _idx_to_id = classes.idx_to_id
                except AttributeError:
                    catnames = [classes[idx] for idx in class_idxs]
                    new_kpts_v2['keypoint_category_name'] = catnames
                else:
                    keypoint_category_id = [_idx_to_id[idx] for idx in class_idxs]
                    new_kpts_v2['keypoint_category_id'] = keypoint_category_id
            visible = self.data.get('visible', None)
            if visible is not None:
                new_kpts_v2['visible'] = self.data['visible'].tolist()
            return new_kpts_v2
        elif style.startswith('new'):
            if style == 'new-id':
                use_id = True
            elif style == 'new-name':
                use_id = False
            elif style == 'new':
                use_id = False
            else:
                raise KeyError(style)

            new_kpts = []
            _visible = self.data.get('visible', None)
            _class_idxs = self.data.get('class_idxs', None)
            for i, xy in enumerate(self.data['xy'].data.tolist()):
                kpdict = {'xy': xy}
                if _visible is not None:
                    kpdict['visible'] = int(_visible[i])
                if _class_idxs is not None:
                    cidx = _class_idxs[i]
                    if use_id:
                        cid = self.meta['classes'].idx_to_id[cidx]
                        kpdict['keypoint_category_id'] = int(cid)
                    else:
                        cname = self.meta['classes'][cidx]
                        kpdict['keypoint_category'] = cname
                new_kpts.append(kpdict)
            return new_kpts
        else:
            raise KeyError(style)

    def to_wkt(self):
        """
        Convert points to well known text. Loses any metadata information.

        Returns:
            str

        Example:
            >>> import kwimage
            >>> self = kwimage.Points.random(3, rng=0)
            >>> self.to_wkt()
        """
        import shapely
        xy = self.data['xy'].data
        geom = shapely.geometry.multipoint.MultiPoint(xy)
        return geom

    def to_shapely(self):
        """
        Convert points to shapely. Loses any metadata information.

        Returns:
            shapely.geometry.multipoint.MultiPoint:

        Example:
            >>> import kwimage
            >>> self = kwimage.Points.random(3)
            >>> geom = self.to_shapely()
            >>> print(f'geom={geom}')
        """
        geom = self.data['xy'].to_shapely()
        return geom

    @classmethod
    def from_shapely(Points, geom):
        """
        Create a Points object from shapely.

        Args:
            geom (shapely.geometry.multipoint.MultiPoint):
                a shapely multipoint

        Returns:
            Self:

        Example:
            >>> import kwimage
            >>> self = kwimage.Points.random(3)
            >>> geom = self.to_shapely()
            >>> new = kwimage.Points.from_shapely(geom)
            >>> assert np.isclose(self.xy, new.xy).all()
        """
        import kwimage
        data = {}
        data['xy'] = kwimage.Coords.from_shapely(geom)
        self = Points(data)
        return self

    @classmethod
    def coerce(cls, data, classes=None):
        """
        Attempt to coerce data into a Points object

        Args:
            data (list | dict | Points): data to attempt to coerce into points
            classes (list[str] | kwcoco.CategoryTree):
                list of all keypoint category names if converting from a
                coco representation.

        Example:
            >>> import kwimage
            >>> data = np.random.rand(10, 2)
            >>> pts = kwimage.Points.coerce(data)
        """
        if isinstance(data, cls):
            return data
        elif isinstance(data, (list, dict)):
            # TODO: determine if coco or geojson
            return cls.from_coco(data, classes=classes)
        elif _generic.isinstance_arraytypes(data):
            return cls(xy=data)
        # TODO: check if a shapely object
        else:
            raise TypeError(type(data))

    @classmethod
    def _from_coco(cls, coco_kpts, class_idxs=None, classes=None):
        # backwards compatibility
        return cls.from_coco(coco_kpts, class_idxs=class_idxs, classes=classes)

    @classmethod
    def from_coco(cls, coco_kpts, class_idxs=None, classes=None, warn=False):
        """
        Args:
            coco_kpts (list | dict): either the original list keypoint encoding
                or the new dict keypoint encoding.

            class_idxs (list): only needed if using old style

            classes (None | list[str] | kwcoco.CategoryTree):
                list of all keypoint category names

            warn (bool): if True raise warnings

        Example:
            >>> ##
            >>> classes = ['mouth', 'left-hand', 'right-hand']
            >>> coco_kpts = [
            >>>     {'xy': (0, 0), 'visible': 2, 'keypoint_category': 'left-hand'},
            >>>     {'xy': (1, 2), 'visible': 2, 'keypoint_category': 'mouth'},
            >>> ]
            >>> Points.from_coco(coco_kpts, classes=classes)
            >>> # Test without classes
            >>> Points.from_coco(coco_kpts)
            >>> # Test without any category info
            >>> coco_kpts2 = [ub.dict_diff(d, {'keypoint_category'}) for d in coco_kpts]
            >>> Points.from_coco(coco_kpts2)
            >>> # Test without category instead of keypoint_category
            >>> coco_kpts3 = [ub.map_keys(lambda x: x.replace('keypoint_', ''), d) for d in coco_kpts]
            >>> Points.from_coco(coco_kpts3)
            >>> #
            >>> # Old style
            >>> coco_kpts = [0, 0, 2, 0, 1, 2]
            >>> Points.from_coco(coco_kpts)
            >>> # Fail case
            >>> coco_kpts4 = [{'xy': [4686.5, 1341.5], 'category': 'dot'}]
            >>> Points.from_coco(coco_kpts4, classes=[])

        Example:
            >>> # xdoctest: +REQUIRES(module:kwcoco)
            >>> import kwcoco
            >>> classes = kwcoco.CategoryTree.from_coco([
            >>>     {'name': 'mouth', 'id': 2}, {'name': 'left-hand', 'id': 3}, {'name': 'right-hand', 'id': 5}
            >>> ])
            >>> coco_kpts = [
            >>>     {'xy': (0, 0), 'visible': 2, 'keypoint_category_id': 5},
            >>>     {'xy': (1, 2), 'visible': 2, 'keypoint_category_id': 2},
            >>> ]
            >>> pts = Points.from_coco(coco_kpts, classes=classes)
            >>> assert pts.data['class_idxs'].tolist() == [2, 0]

        Example:
            >>> # xdoctest: +REQUIRES(module:kwcoco)
            >>> from kwimage.structs.points import *  # NOQA
            >>> import kwcoco
            >>> coco_kpts = {
            >>>     'x': [0, 1, 2, 3, 4, 5],
            >>>     'y': [0, 1, 2, 3, 4, 5],
            >>> }
            >>> # Without classes
            >>> pts1 = Points.from_coco(coco_kpts)
            >>> print(f'pts1.data={pts1.data}')
            >>> print(f'pts1.meta={pts1.meta}')
            >>> # With classes
            >>> classes = kwcoco.CategoryTree.from_coco([
            >>>     {'name': 'mouth', 'id': 2}, {'name': 'left-hand', 'id': 3}, {'name': 'right-hand', 'id': 5}
            >>> ])
            >>> coco_kpts['keypoint_category_id'] = [5, 2, 3, 5, 2]
            >>> pts2 = Points.from_coco(coco_kpts, classes=classes)
            >>> print(f'pts2 = {ub.urepr(pts2, nl=1)}')
            >>> print(f'pts2.data={pts2.data}')
            >>> print(f'pts2.meta={pts2.meta}')
            >>> assert pts2.data['class_idxs'].tolist() == [2, 0, 1, 2, 0]
        """
        if coco_kpts is None:
            return None

        if isinstance(coco_kpts, dict) and 'x' in coco_kpts:
            # new style (v2) which is column based and uses less memory
            xs = coco_kpts['x']
            ys = coco_kpts['y']
            class_idxs = None
            visible = None
            if len(xs) != len(ys):
                raise ValueError(
                    'new-v2 column-style coco keypoints must have '
                    f'corresponding length lists. Got: {len(xs)}, {len(ys)}'
                )

            visible = coco_kpts.get('visible', None)
            if visible is None:
                visible = np.array([2] * len(xs))
            else:
                visible = np.array(visible)

            keypoint_category_id = coco_kpts.get('keypoint_category_id', None)
            if keypoint_category_id is not None:
                if classes is None:
                    class_idxs = None
                    # raise Exception('classes should be specified for new-style-v2')
                else:
                    try:
                        class_idxs = [classes.id_to_idx[cid]
                                      for cid in keypoint_category_id]
                    except AttributeError:
                        raise TypeError('classes needs to be a kwcoco.CategoryTree to parse keypoint_category_id')
                    else:
                        class_idxs = np.array(class_idxs)

            xs = np.array(xs)
            ys = np.array(ys)
            xy = np.stack([xs, ys], axis=1)
            self = cls(xy=xy, visible=visible, class_idxs=class_idxs,
                       classes=classes)

        elif len(coco_kpts) and isinstance(ub.peek(coco_kpts), dict):
            # new style (v1)
            xy = []
            visible = []
            cidx_list = []

            if class_idxs is not None:
                if warn:
                    warnings.warn('class_idxs should not be specified for new-style')
                class_idxs = None

            drop_cidx = False

            # raise NotImplementedError(
            #     '''
            #     Needs to have extra information available to map
            #     between keypoint category ids and idxs.
            #     ''')

            if classes is None or not bool(classes):
                # See if we can infer the classes.
                # This may cause compatiblity issues.
                inferred_classes = [kpdict.get('keypoint_category',
                                               kpdict.get('category', None))
                                    for kpdict in coco_kpts]
                if all(inferred_classes):
                    if warn:
                        warnings.warn(
                            'Inferring keypoint classes in Points.from_coco. '
                            'It would be better to specify them explicitly')
                    classes = sorted(set(inferred_classes))

            for kpdict in coco_kpts:
                if classes is not None:
                    if 'keypoint_category_id' in kpdict:
                        cid = kpdict['keypoint_category_id']
                        try:
                            cidx = classes.id_to_idx[cid]
                        except AttributeError:
                            raise TypeError('classes needs to be a kwcoco.CategoryTree to parse keypoint_category_id')
                    elif 'keypoint_category' in kpdict:
                        assert classes is not None
                        cname = kpdict['keypoint_category']
                        cidx = classes.index(cname)
                    elif 'category_name' in kpdict:
                        assert classes is not None
                        cname = kpdict['category_name']
                        cidx = classes.index(cname)
                    ### Legacy support, these are not prefered names ###
                    elif 'category_id' in kpdict:
                        if warn:
                            warnings.warn('Keypoints got category_id, but we would prefer keypoint_category_id')
                        cid = kpdict['category_id']
                        try:
                            cidx = classes.id_to_idx[cid]
                        except AttributeError:
                            raise TypeError('classes needs to be a kwcoco.CategoryTree to parse keypoint_category_id')
                    elif 'category' in kpdict:
                        if warn:
                            warnings.warn('Keypoints got category, but we would prefer keypoint_category')
                        assert classes is not None
                        cname = kpdict['category']
                        cidx = classes.index(cname)
                    # else:
                    #     raise Exception('Keypoint category was not specified')
                    cidx_list.append(cidx)
                else:
                    if 'keypoint_category_id' in kpdict or 'keypoint_category' in kpdict:
                        # warnings.warn('classes should be specified for new-style')
                        # raise Exception('classes should be specified for new-style')
                        cidx_list.append(None)
                        drop_cidx = True

                xy.append(kpdict['xy'])
                visible.append(kpdict.get('visible', 2))

            if cidx_list:
                assert len(cidx_list) == len(xy), 'missing category indices'
            else:
                cidx_list = None

            if drop_cidx:
                cidx_list = None
            else:
                cidx_list = np.array(cidx_list)

            xy = np.array(xy)
            visible = np.array(visible)
            self = cls(xy=xy, visible=visible, class_idxs=cidx_list,
                       classes=classes)
        else:
            # original style
            kp = np.array(coco_kpts).reshape(-1, 3)
            xy = kp[:, 0:2]
            visible = kp[:, 2]
            if class_idxs is not None:
                if len(class_idxs) == 0:
                    if len(kp) > 0:
                        if warn:
                            warnings.warn('Creating keypoints with unknown class information')
                        # raise Exception('Creating keypoints with unknown class information')
                        class_idxs = [-1] * len(xy)
                    else:
                        class_idxs = []
                else:
                    assert len(class_idxs) == len(xy), '{} {}'.format(
                        len(class_idxs), len(xy))
            self = cls(xy=xy, visible=visible, class_idxs=class_idxs,
                       classes=classes)
        return self


class PointsList(_generic.ObjectList):
    """
    Stores a list of Points, each item usually corresponds to a different object.

    Note:
        # TODO: when the data is homogenous we can use a more efficient
        # representation, otherwise we have to use heterogenous storage.
    """



def test_rational_affine():
    from kwimage import Affine
    import numbers
    import ubelt as ub
    import numpy as np
    import kwarray
    import pytest

    try:
        import sympy
    except ImportError:
        pytest.skip()

    rng = kwarray.ensure_rng(None)
    offset = [0, 0]
    about = [0, 0]
    theta = 0
    shearx = 0
    scale = rng.randn(2) * 10
    offset = rng.randn(2) * 10
    about = rng.randn(2) * 10
    theta = rng.randn() * 10
    shearx = rng.randn() * 10

    def as_rational(data):
        if isinstance(data, list):
            rat = list(map(as_rational, data))
        elif isinstance(data, numbers.Number):
            rat = sympy.Rational(data)
        elif isinstance(data, np.ndarray):
            flat_rat = list(map(sympy.Rational, data.ravel().tolist()))
            rat = sympy.Array(flat_rat).reshape(*data.shape)
        elif isinstance(data, sympy.Rational):
            rat = data
        elif isinstance(data, sympy.Array):
            # TODO: check rationality
            rat = data
        else:
            raise TypeError(type(data))
        return rat

    scale = as_rational(scale)
    offset = as_rational(offset)
    about = as_rational(about)
    theta = as_rational(theta)
    shearx = as_rational(shearx)

    math_mod = sympy
    array_cls = sympy.Matrix

    # Create combined matrix from all params
    A = Affine.affine(
        scale=scale, offset=offset, theta=theta, shearx=shearx,
        about=about, array_cls=array_cls, math_mod=math_mod)
    # Test that combining components matches
    S = Affine.affine(scale=scale, array_cls=array_cls, math_mod=math_mod)
    T = Affine.affine(offset=offset, array_cls=array_cls, math_mod=math_mod)
    R = Affine.affine(theta=theta, array_cls=array_cls, math_mod=math_mod)
    H = Affine.affine(shearx=shearx, array_cls=array_cls, math_mod=math_mod)
    F = Affine.affine(offset=about, array_cls=array_cls, math_mod=math_mod)
    # combine (note shear must be on the RHS of rotation)
    alt  = F @ T @ R @ H @ S @ F.inv()

    V2 = np.asarray(alt.matrix).astype(float)
    V1 = np.asarray(A.matrix).astype(float)
    print(f'V2={V2}')
    print(f'V1={V1}')

    print('A    = {}'.format(ub.urepr(A.matrix.tolist(), nl=1)))
    print('alt  = {}'.format(ub.urepr(alt.matrix.tolist(), nl=1)))
    assert np.all(V1 == V2)
    pt = np.vstack([np.random.rand(2, 1), [[1]]])
    warp_pt1 = (A.matrix @ pt).evalf()
    warp_pt2 = (alt.matrix @ pt).evalf()
    assert warp_pt2 == warp_pt1


def test_duplicate_points_non_rank_defficient():
    import kwimage
    import numpy as np
    pts1 = np.array([
        [1, 1],
        [1, 10],
        [10, 1],
        [10, 1],
        [10, 1],
        [10, 10],
    ])
    pts2 = np.array([
        [1, 1],
        [1, 10],
        [10, 1],
        [10, 1],
        [10, 1],
        [10, 10],
    ])
    aff = kwimage.Affine.fit(pts1, pts2)
    pts2_recon = kwimage.Points(xy=pts1).warp(aff)
    assert np.allclose(pts2_recon.xy, pts2)


def test_rank_defficient_without_duplicates():
    import kwimage
    import numpy as np
    pts1 = np.array([
        [1, 1],
        [10, 10],
    ])
    pts2 = np.array([
        [1, 1],
        [10, 10],
    ])
    aff = kwimage.Affine.fit(pts1, pts2)
    pts2_recon = kwimage.Points(xy=pts1).warp(aff)
    assert np.allclose(pts2_recon.xy, pts2)


def test_rank_defficient_with_duplicates():
    import kwimage
    import numpy as np
    pts1 = np.array([
        [8, 4],
        [8, 4],
        [9, 8],
        [9, 8],
    ])
    pts2 = np.array([
        [8, 4],
        [8, 4],
        [9, 8],
        [9, 8],
    ])
    aff = kwimage.Affine.fit(pts1, pts2)
    pts2_recon = kwimage.Points(xy=pts1).warp(aff)
    assert np.allclose(pts2_recon.xy, pts2)


def test_rank_defficient_with_duplicates_non_exact():
    import kwimage
    import numpy as np
    pts1 = np.array([[200.244, 250.543],
                     [200.247, 250.545],
                     [100.754, 0.265],
                     [100.754, 0.265]])

    pts2 = np.array([[200.25, 250.54],
                     [200.25, 250.54],
                     [100.77, 0.26],
                     [100.77, 0.26]])

    aff = kwimage.Affine.fit(pts1, pts2)
    pts2_recon = kwimage.Points(xy=pts1).warp(aff)
    assert np.allclose(pts2_recon.xy, pts2)


def test_single_correspondence():
    import kwimage
    import numpy as np
    pts1 = np.array([
        [1, 1],
    ])
    pts2 = np.array([
        [10, 10],
    ])
    aff = kwimage.Affine.fit(pts1, pts2)
    pts2_recon = kwimage.Points(xy=pts1).warp(aff)
    assert np.allclose(pts2_recon.xy, pts2)



"""
Fast R-CNN
Copyright (c) 2015 Microsoft
Licensed under The MIT License [see LICENSE for details]
Written by Ross Girshick
"""
import numpy as np
import warnings


def py_nms(np_ltrb, np_scores, thresh, bias=1):
    """
    Pure Python NMS baseline.

    References:
        https://github.com/rbgirshick/fast-rcnn/blob/master/lib/utils/nms.py

    CommandLine:
        xdoctest -m kwimage.algo._nms_backend.py_nms py_nms

    Example:
        >>> from kwimage.algo._nms_backend.py_nms import *  # NOQA
        >>> np_ltrb = np.array([
        >>>     [0, 0, 100, 100],
        >>>     [0, 0, 100, 100],
        >>>     [100, 100, 10, 10],
        >>>     [10, 10, 100, 100],
        >>>     [50, 50, 100, 100],
        >>>     [100, 100, 150, 101],
        >>>     [120, 100, 180, 101],
        >>>     [150, 100, 200, 101],
        >>> ], dtype=np.float32)
        >>> np_scores = np.linspace(0, 1, len(np_ltrb))
        >>> thresh = 0.1
        >>> bias = 0.0
        >>> keep = sorted(map(int, py_nms(np_ltrb, np_scores, thresh, bias)))
        >>> print('keep = {!r}'.format(keep))
        keep = [2, 4, 5, 7]

    Example:
        >>> from kwimage.algo._nms_backend.py_nms import *  # NOQA
        >>> np_ltrb = np.array([
        >>>     [0, 0, 100, 100],
        >>>     [100, 100, 10, 10],
        >>>     [10, 10, 100, 100],
        >>>     [50, 50, 100, 100],
        >>> ], dtype=np.float32)
        >>> np_scores = np.array([.1, .5, .9, .1])
        >>> keep = list(map(int, py_nms(np_ltrb, np_scores, thresh=0.0, bias=1.0)))
        >>> print('keep@0.0 = {!r}'.format(keep))
        >>> keep = list(map(int, py_nms(np_ltrb, np_scores, thresh=0.2, bias=1.0)))
        >>> print('keep@0.2 = {!r}'.format(keep))
        >>> keep = list(map(int, py_nms(np_ltrb, np_scores, thresh=0.5, bias=1.0)))
        >>> print('keep@0.5 = {!r}'.format(keep))
        >>> keep = list(map(int, py_nms(np_ltrb, np_scores, thresh=1.0, bias=1.0)))
        >>> print('keep@1.0 = {!r}'.format(keep))
        keep@0.0 = [2, 1]
        keep@0.2 = [2, 1]
        keep@0.5 = [2, 1, 3]
        keep@1.0 = [2, 1, 3, 0]

    Example:
        >>> # Test int16 case
        >>> from kwimage.algo._nms_backend.py_nms import *  # NOQA
        >>> np_ltrb = np.array([
        >>>    [  71.75,  -37.25,  373.8 ,  231.2 ],
        >>>    [ 282.  ,  395.5 ,  609.  ,  639.5 ],
        >>>    [  36.62,   -5.5 ,  386.  ,  321.8 ],
        >>>    [ 207.5 ,  546.5 ,  238.6 ,  563.  ]], dtype=np.float16)
        >>> np_scores = np.linspace(0, 1, len(np_ltrb))
        >>> thresh = 0.1
        >>> bias = 0.0
        >>> keep = sorted(map(int, py_nms(np_ltrb, np_scores, thresh, bias)))
        >>> print('keep = {!r}'.format(keep))
    """
    new_dtype = np.result_type(np_ltrb, np.float32)
    if new_dtype != np_ltrb.dtype:
        # Upcast if precision is too low
        np_ltrb = np_ltrb.astype(new_dtype)

    x1 = np_ltrb[:, 0]
    y1 = np_ltrb[:, 1]
    x2 = np_ltrb[:, 2]
    y2 = np_ltrb[:, 3]

    widths = (x2 - x1 + bias)
    heights = (y2 - y1 + bias)
    areas = widths * heights

    idxs_remain = np_scores.argsort()[::-1]

    keep = []

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'invalid value .* true_divide')

        # n_conflicts = 0
        while idxs_remain.size > 0:
            i = idxs_remain[0]
            keep.append(i)
            # print('Keeping the i={}-th box'.format(i))

            idxs_remain = idxs_remain[1:]
            # print('Compute IoU between chosen and remaining boxes')
            xx1 = np.maximum(x1[i], x1[idxs_remain])
            yy1 = np.maximum(y1[i], y1[idxs_remain])
            xx2 = np.minimum(x2[i], x2[idxs_remain])
            yy2 = np.minimum(y2[i], y2[idxs_remain])

            w = np.maximum(0.0, xx2 - xx1 + bias)
            h = np.maximum(0.0, yy2 - yy1 + bias)
            inter = w * h
            denom = (areas[i] + areas[idxs_remain] - inter)
            iou = inter / denom
            iou = np.nan_to_num(iou)
            # print('Checking overlap between the i={}-th and remaining boxes'.format(i))

            # Keep anything that doesnt have a large overlap with this item
            flags = iou <= thresh
            inds = np.where(flags)[0]
            # print('Supress {} boxes that did overlap this one'.format(len(ovr) - len(inds)))
            # print('Consider {} boxes that dont overlap this one'.format(len(inds)))
            idxs_remain = idxs_remain[inds]
    return keep


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m kwimage.algo._nms_backend.py_nms
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



from typing import Any
from typing import Tuple
from typing import Callable
import kwimage
from numpy.typing import ArrayLike
from numpy import ndarray
from typing import List
from typing import Dict
import kwcoco
from _typeshed import Incomplete
from kwimage.structs import _generic

from kwimage._typing import SKImageGeometricTransform

__docstubs__: str


class _PointsWarpMixin:

    def to_imgaug(self, input_dims):
        ...

    @classmethod
    def from_imgaug(cls, kpoi):
        ...

    @property
    def dtype(self):
        ...

    def warp(self,
             transform: ArrayLike | Callable | kwimage.Affine
             | SKImageGeometricTransform | Any,
             input_dims: Tuple | None = None,
             output_dims: Tuple | None = None,
             inplace: bool = False):
        ...

    def scale(self,
              factor: float | Tuple[float, float],
              output_dims: Tuple | None = None,
              inplace: bool = ...):
        ...

    def translate(self,
                  offset,
                  output_dims: Tuple | None = None,
                  inplace: bool = ...):
        ...


class Points(_generic.Spatial, _PointsWarpMixin):
    __datakeys__: Incomplete
    __metakeys__: Incomplete
    data: Incomplete
    meta: Incomplete

    def __init__(self,
                 data: Incomplete | None = ...,
                 meta: Incomplete | None = ...,
                 datakeys: Incomplete | None = ...,
                 metakeys: Incomplete | None = ...,
                 **kwargs) -> None:
        ...

    def __nice__(self):
        ...

    def __len__(self):
        ...

    @property
    def shape(self):
        ...

    @property
    def xy(self):
        ...

    @classmethod
    def random(Points,
               num: int = ...,
               classes: Incomplete | None = ...,
               rng: Incomplete | None = ...):
        ...

    def is_numpy(self):
        ...

    def is_tensor(self):
        ...

    def tensor(self, device=...):
        ...

    def round(self, inplace: bool = False):
        ...

    def numpy(self):
        ...

    def draw_on(self,
                image: ndarray | None = None,
                color: str | Any | List[Any] = 'white',
                radius: None | int = None,
                copy: bool = False):
        ...

    def draw(self,
             color: str = ...,
             ax: Incomplete | None = ...,
             alpha: Incomplete | None = ...,
             radius: int = ...,
             setlim: bool = ...,
             **kwargs):
        ...

    def compress(self, flags, axis: int = ..., inplace: bool = ...):
        ...

    def take(self, indices, axis: int = ..., inplace: bool = ...):
        ...

    @classmethod
    def concatenate(cls, points, axis: int = ...):
        ...

    def to_coco(self, style: str = 'orig') -> Dict:
        ...

    @classmethod
    def coerce(cls, data):
        ...

    @classmethod
    def from_coco(cls,
                  coco_kpts: list | dict,
                  class_idxs: list | None = None,
                  classes: list | kwcoco.CategoryTree | None = None,
                  warn: bool = False):
        ...


class PointsList(_generic.ObjectList):
    ...



"""
Helpers to query information about available backends
"""

try:
    from functools import cache
except ImportError:
    from ubelt import memoize as cache


@cache
def _have_turbojpg():
    """
    pip install PyTurboJPEG

    """
    try:
        import turbojpeg  # NOQA
        turbojpeg.TurboJPEG()
    except Exception:
        return False
    else:
        return True


@cache
def _have_gdal():
    try:
        from osgeo import gdal  # NOQA
    except Exception:
        return False
    else:
        return True


@cache
def _have_cv2():
    try:
        import cv2  # NOQA
    except Exception:
        return False
    else:
        return True


@cache
def _default_backend():
    """
    Define the default backend for simple cases.
    In kwimage < 0.11.0, this was always cv2, but now cv2 is optional, so we
    will fallback to skimage (or is PIL a better option?)
    """
    if _have_cv2():
        return 'cv2'
    else:
        return 'skimage'



import kwimage
import ubelt as ub
import numpy as np


def _random_drawables():
    from kwimage._backend_info import _have_cv2
    if not _have_cv2():
        import pytest
        pytest.skip('requires cv2')
    drawables = {
        'boxes': kwimage.Boxes.random(),
        'points': kwimage.Points.random(),
        'polys': kwimage.Polygon.random(),
        'mpolys': kwimage.MultiPolygon.random(),
        'dets': kwimage.Polygon.random(),
    }
    return drawables


def _random_extreme_images():
    images = {}
    for w in [0, 1, 2]:
        for h in [0, 1, 2]:
            for nchans in [None, 1, 3]:
                key = f'{w}x{h}x{nchans}'
                if nchans is None:
                    shape = (h, w)
                else:
                    shape = (h, w, nchans)
                data = np.random.rand(*shape)
                images[key] = data
    return images


def test_draw_on_extreme_sizes():
    """
    Test that draw_on doesn't fail with zero sized images or small images in
    general.
    """
    from kwimage._backend_info import _have_cv2
    if not _have_cv2():
        import pytest
        pytest.skip('requires cv2')
    drawables = _random_drawables()
    images = _random_extreme_images()

    errors = []
    for draw_key, drawable in drawables.items():
        for im_key, image in images.items():

            try:
                result = drawable.draw_on(image)
                assert result.shape[0:2] == image.shape[0:2]
            except Exception as ex:
                errors.append((draw_key, im_key, ex))
                raise

    if errors:
        print('errors = {}'.format(ub.urepr(errors, nl=1)))
        raise AssertionError('errors = {}'.format(ub.urepr(errors, nl=1)))



"""
Functions for stacking images (of potentially different sizes) together in a
single image.

Notes:
    * We may change the "bg_value" argument to "bg_color" in the future.
"""

import cv2
import numpy as np
import skimage
import skimage.transform
from . import im_core
from . import im_cv2


def stack_images(images, axis=0, resize=None, interpolation=None, overlap=0,
                 return_info=False, bg_value=None, pad=None,
                 allow_casting=True):
    """
    Make a new image with the input images side-by-side

    Args:
        images (Iterable[ndarray]):  image data

        axis (int): axis to stack on (either 0 or 1)

        resize (int | str | None):
            if None image sizes are not modified, otherwise resize resize can
            be either 0 or 1.  We resize the `resize`-th image to match the
            `1 - resize`-th image. In other words, resize=0 means the current
            image is resized to match the next image, and resize=1 means the
            next image is resized to match the current image.
            Can also be strings "larger" or "smaller". If resize="larger", the
            current image is only resized if the next image is larger. If
            resize='smaller' the current image is resized only if the next
            image is smaller.
            TODO: this needs to be reworked to be more intuitive.

        interpolation (int | str):
            string or cv2-style interpolation type.
            only used if resize or overlap > 0

        overlap (int):
            number of pixels to overlap. Using a negative number results in a
            border.

        pad (int):
            if specified overrides `overlap` as a the number of pixels to pad
            between images.

        return_info (bool):
            if True, returns transforms (scales and translations) to map from
            original image to its new location.

        bg_value (Number | ndarray | str):
            background value or color, if specified, uses this as a fill value.

        allow_casting (bool):
            if True, then if "uint255" and "float01" format images are given
            they are converted to "float01".  Defaults to True.

    Returns:
        ndarray: an image of stacked images side by side

        Tuple[ndarray, List]: where the first item is the aformentioned stacked
            image and the second item is a list of transformations for each
            input image mapping it to its location in the returned image.

    SeeAlso:
        :func:`kwimage.im_stack.stack_images_grid`

    TODO:
        - [ ] This is currently implemented by calling the "stack_two_images"
              function multiple times. This should be optimized by allocating
              the entire canvas first.

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import kwimage
        >>> img1 = kwimage.grab_test_image('carl', space='rgb')
        >>> img2 = kwimage.grab_test_image('astro', space='rgb')
        >>> images = [img1, img2]
        >>> imgB, transforms = kwimage.stack_images(
        >>>     images, axis=0, resize='larger', pad=10, return_info=True)
        >>> print('imgB.shape = {}'.format(imgB.shape))
        >>> # xdoctest: +REQUIRES(--show)
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import kwplot
        >>> import kwimage
        >>> kwplot.autompl()
        >>> kwplot.imshow(imgB, colorspace='rgb')
        >>> wh1 = np.multiply(img1.shape[0:2][::-1], transforms[0].scale)
        >>> wh2 = np.multiply(img2.shape[0:2][::-1], transforms[1].scale)
        >>> xoff1, yoff1 = transforms[0].translation
        >>> xoff2, yoff2 = transforms[1].translation
        >>> xywh1 = (xoff1, yoff1, wh1[0], wh1[1])
        >>> xywh2 = (xoff2, yoff2, wh2[0], wh2[1])
        >>> kwplot.draw_boxes(kwimage.Boxes([xywh1], 'xywh'), color=(1.0, 0, 0))
        >>> kwplot.draw_boxes(kwimage.Boxes([xywh2], 'xywh'), color=(1.0, 0, 0))
        >>> kwplot.show_if_requested()
    """
    imgiter = iter(images)
    try:
        img1 = next(imgiter)
    except StopIteration:
        raise ValueError('need at least one image to stack')

    if pad is not None:
        overlap = -pad

    if return_info:
        transforms_ = [skimage.transform.AffineTransform(
            scale=[1.0, 1.0], translation=[0.0, 0.0]
        )]

    for img2 in imgiter:
        img1, offset_tup, sf_tup = _stack_two_images(img1, img2, axis=axis,
                                                     resize=resize,
                                                     bg_value=bg_value,
                                                     interpolation=interpolation,
                                                     overlap=overlap,
                                                     allow_casting=allow_casting)
        if return_info:
            off1, off2 = offset_tup
            sf1, sf2 = sf_tup

            tf1 = skimage.transform.AffineTransform(scale=sf1, translation=off1)
            tf2 = skimage.transform.AffineTransform(scale=sf2, translation=off2)

            # Apply transforms to the first image to all previous transforms
            for t in transforms_:
                t.params = t.params.dot(tf1.params)
            # Append the second transform
            transforms_.append(tf2)

    if return_info:
        return img1, transforms_
    else:
        return img1


def stack_images_grid(images, chunksize=None, axis=0, overlap=0, pad=None,
                      return_info=False, bg_value=None, resize=None,
                      allow_casting=True):
    """
    Stacks images in a grid. Optionally return transforms of original image
    positions in the output image.

    Args:
        images (Iterable[ndarray]):  image data

        chunksize (int): number of rows per column or columns per
            row depending on the value of `axis`.
            If unspecified, computes this as `int(sqrt(len(images)))`.

        axis (int):
            If 0, chunksize is columns per row.
            If 1, chunksize is rows per column.
            Defaults to 0.

        overlap (int): number of pixels to overlap. Using a negative
            number results in a border.

        pad (int): if specified overrides `overlap` as a the number of pixels
            to pad between images.

        return_info (bool): if True, returns transforms (scales and
            translations) to map from original image to its new location.

        resize (int | str | None):
            if None image sizes are not modified, otherwise can be set to
            "larger" or "smaller" to resize the images in each stack direction.

        bg_value (Number | ndarray | str):
            background value or color, if specified, uses this as a fill value.

        allow_casting (bool):
            if True, then if "uint255" and "float01" format images are given
            they are converted to "float01".  Defaults to True.

    Returns:
        ndarray: an image of stacked images in a grid pattern

        Tuple[ndarray, List]: where the first item is the aformentioned stacked
            image and the second item is a list of transformations for each
            input image mapping it to its location in the returned image.

    SeeAlso:
        :func:`kwimage.im_stack.stack_images`

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import kwimage
        >>> img1 = kwimage.grab_test_image('carl')
        >>> img2 = kwimage.grab_test_image('astro')
        >>> img3 = kwimage.grab_test_image('airport')
        >>> img4 = kwimage.grab_test_image('paraview')[..., 0:3]
        >>> img5 = kwimage.grab_test_image('pm5644')
        >>> images = [img1, img2, img3, img4, img5]
        >>> canvas, transforms = kwimage.stack_images_grid(
        ...     images, chunksize=3, axis=0, pad=10, bg_value='kitware_blue',
        ...     return_info=True, resize='larger')
        >>> print('canvas.shape = {}'.format(canvas.shape))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> import kwimage
        >>> kwplot.autompl()
        >>> kwplot.imshow(canvas)
        >>> kwplot.show_if_requested()
    """
    import ubelt as ub
    if chunksize is None:
        chunksize = int(len(images) ** .5)
    if pad is not None:
        overlap = -pad

    if chunksize == 0:
        raise ValueError('need at least one image to stack and chunksize cannot be zero')

    stack1_list = []
    tfs1_list = []
    assert axis in [0, 1]
    try:
        for batch in ub.chunks(images, chunksize, bordermode='none'):
            stack1, tfs1 = stack_images(batch, overlap=overlap, return_info=True,
                                        bg_value=bg_value,
                                        resize=resize, axis=1 - axis)
            tfs1_list.append(tfs1)
            stack1_list.append(stack1)
    except ZeroDivisionError:
        raise ValueError('need at least one image to stack')

    img_grid, tfs2 = stack_images(stack1_list, overlap=overlap,
                                  bg_value=bg_value,
                                  return_info=True, axis=axis)
    transforms_ = [tf1 + tf2
                   for tfs1, tf2 in zip(tfs1_list, tfs2)
                   for tf1 in tfs1]

    if return_info:
        return img_grid, transforms_
    else:
        return img_grid


def _stack_two_images(img1, img2, axis=0, resize=None, interpolation=None,
                      overlap=0, bg_value=None, allow_casting=True):
    """
    Returns:
        Tuple[ndarray, Tuple, Tuple]: imgB, offset_tup, sf_tup

    Ignore:
        import xinspect
        globals().update(xinspect.get_func_kwargs(_stack_two_images))
        resize = 1
        overlap = -10
    """

    def _rectify_axis(img1, img2, axis):
        """ determine if we are stacking in horzontally or vertically """
        (h1, w1) = img1.shape[0: 2]  # get chip dimensions
        (h2, w2) = img2.shape[0: 2]
        xoff2, yoff2 = 0, 0
        vert_wh  = max(w1, w2), h1 + h2
        horiz_wh = w1 + w2, max(h1, h2)
        if axis is None:
            # Display the orientation with the better (closer to 1) aspect ratio
            vert_ar  = max(vert_wh) / min(vert_wh)
            horiz_ar = max(horiz_wh) / min(horiz_wh)
            axis = 0 if vert_ar < horiz_ar else 1
        if axis == 0:  # vertical stack
            wB, hB = vert_wh
            yoff2 = h1
        elif axis == 1:
            wB, hB = horiz_wh
            xoff2 = w1
        else:
            raise ValueError('axis can only be 0 or 1')
        return axis, h1, h2, w1, w2, wB, hB, xoff2, yoff2

    def _round_dsize(dsize, scale):
        """
        Returns an integer size and scale that best approximates
        the floating point scale on the original size

        Args:
            dsize (tuple): original width height
            scale (float | tuple): desired floating point scale factor
        """
        try:
            sx, sy = scale
        except TypeError:
            sx = sy = scale
        w, h = dsize
        new_w = int(round(w * sx))
        new_h = int(round(h * sy))
        new_scale = new_w / w, new_h / h
        new_dsize = (new_w, new_h)
        return new_dsize, new_scale

    def _ramp(shape, axis):
        """ nd ramp function """
        newshape = [1] * len(shape)
        reps = list(shape)
        newshape[axis] = -1
        reps[axis] = 1
        basis = np.linspace(0, 1, shape[axis])
        data = basis.reshape(newshape)
        return np.tile(data, reps)

    def _blend(part1, part2, alpha):
        """ blending based on an alpha mask """
        part1, alpha = im_core.make_channels_comparable(part1, alpha)
        part2, alpha = im_core.make_channels_comparable(part2, alpha)
        partB = (part1 * (1.0 - alpha)) + (part2 * (alpha))
        return partB

    interpolation = im_cv2._coerce_interpolation(interpolation,
                                                 default=cv2.INTER_NEAREST)

    if allow_casting:
        if img1.dtype.kind == 'f' and img2.dtype.kind == 'u':
            img2 = im_core.ensure_float01(img2)
        elif img1.dtype.kind == 'u' and img2.dtype.kind == 'f':
            img1 = im_core.ensure_float01(img1)

        if img1.dtype != img2.dtype:
            if img1.dtype.kind == 'f' and img2.dtype.kind == 'f':
                common_dtype = np.promote_types(img1.dtype, img2.dtype)
                img1 = img1.astype(common_dtype)
                img2 = img2.astype(common_dtype)

    img1, img2 = im_core.make_channels_comparable(img1, img2)
    nChannels = im_core.num_channels(img1)

    if img1.dtype != img2.dtype:
        raise TypeError(
            'img1.dtype=%r, img2.dtype=%r' % (img1.dtype, img2.dtype))

    axis, h1, h2, w1, w2, wB, hB, xoff2, yoff2 = _rectify_axis(img1, img2, axis)

    # Rectify both images to they are the same dimension
    if resize:
        # Compre the lengths of the width and height
        length1 = img1.shape[1 - axis]
        length2 = img2.shape[1 - axis]

        if resize == 'larger':
            resize = 0 if length1 > length2 else 1
        elif resize == 'smaller':
            resize = 0 if length1 < length2 else 1

        if resize == 0:
            # Resize the first image
            sf2 = (1., 1.)
            scale = length2 / length1
            dsize, sf1 = _round_dsize(img1.shape[0:2][::-1], scale)
            img1 = cv2.resize(img1, dsize, interpolation=interpolation)
        elif resize == 1:
            # Resize the second image
            sf1 = (1., 1.)
            scale = length1 / length2
            dsize, sf2 = _round_dsize(img2.shape[0:2][::-1], scale)
            img2 = cv2.resize(img2, dsize, interpolation=interpolation)
        else:
            raise ValueError('resize can only be 0 or 1 - or a special key')
        axis, h1, h2, w1, w2, wB, hB, xoff2, yoff2 = _rectify_axis(img1, img2, axis)
    else:
        sf1 = (1.0, 1.0)
        sf2 = (1.0, 1.0)

    # allow for some overlap / blending of the images
    if overlap != 0:
        if axis == 0:
            hB -= overlap
        elif axis == 1:
            wB -= overlap

    # Do image concatentation
    if nChannels > 1 or len(img1.shape) > 2:
        newshape = (hB, wB, nChannels)
    else:
        newshape = (hB, wB)
    # Allocate new image for both
    imgB = np.zeros(newshape, dtype=img1.dtype)

    if bg_value is not None:
        if isinstance(bg_value, str):
            import kwimage
            bg_value = kwimage.Color(bg_value).forimage(imgB)
        try:
            imgB[:, :] = bg_value
        except ValueError:
            imgB = im_core.atleast_3channels(imgB)
            imgB[:, :] = bg_value

    # Insert the images in the larger frame

    # Insert the first image
    xoff1 = yoff1 = 0
    imgB[yoff1:(yoff1 + h1), xoff1:(xoff1 + w1)] = img1

    if overlap:
        if axis == 0:
            yoff2 -= overlap
        elif axis == 1:
            xoff2 -= overlap

        imgB[yoff2:(yoff2 + h2), xoff2:(xoff2 + w2)] = img2

        if overlap > 0:
            # Blend the overlapping part
            if axis == 0:
                part1 = img1[-overlap:, :]
                part2 = imgB[yoff2:(yoff2 + overlap), 0:w1]
                alpha = _ramp(part1.shape[0:2], axis=axis)
                blended = _blend(part1, part2, alpha)
                imgB[yoff2:(yoff2 + overlap), 0:w1] = blended
            elif axis == 1:
                part1 = img1[:, -overlap:]
                part2 = imgB[0:h1, xoff2:(xoff2 + overlap)]
                alpha = _ramp(part1.shape[0:2], axis=axis)
                blended = _blend(part1, part2, alpha)
                imgB[0:h1, xoff2:(xoff2 + overlap)] = blended
    else:
        imgB[yoff2:(yoff2 + h2), xoff2:(xoff2 + w2)] = img2

    offset1 = (xoff1, yoff1)
    offset2 = (xoff2, yoff2)
    offset_tup = (offset1, offset2)
    sf_tup = (sf1, sf2)
    return imgB, offset_tup, sf_tup


def _efficient_rectangle_packing():
    """
    References:
        https://en.wikipedia.org/wiki/Packing_problems
        https://github.com/Penlect/rectangle-packer
        https://github.com/secnot/rectpack
        https://stackoverflow.com/questions/1213394/what-algorithm-can-be-used-for-packing-rectangles-of-different-sizes-into-the-sm
        https://www.codeproject.com/Articles/210979/Fast-optimizing-rectangle-packing-algorithm-for-bu

    Requires:
        pip install rectangle-packer

    Ignore:
        >>> import kwimage
        >>> anchors = anchors=[[1, 1], [3 / 4, 1], [1, 3 / 4]]
        >>> boxes = kwimage.Boxes.random(num=100, anchors=anchors).scale((100, 100)).to_xywh()
        >>> # Create a bunch of rectangles (width, height)
        >>> sizes = boxes.data[:, 2:4].astype(int).tolist()
        >>> import rpack
        >>> positions = rpack.pack(sizes)
        >>> boxes.data[:, 0:2] = positions
        >>> boxes = boxes.scale(0.95, about='center')
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> boxes.draw()
        >>> # The result will be a list of (x, y) positions:
        >>> positions

        images = [kwimage.grab_test_image(key) for key in kwimage.grab_test_image.keys()]
        images = [kwimage.imresize(g, max_dim=256) for g in images]

        sizes = [g.shape[0:2][::-1] for g in images]

        import rpack
        positions = rpack.pack(sizes)

        !pip install rectpack

        import rectpack

        bin_width = 512

        packer = rectpack.newPacker(rotation=False)
        for rid, (w, h) in enumerate(sizes):
            packer.add_rect(w, h, rid=rid)

        max_w, max_h = np.array(sizes).sum(axis=0)
        # f = max_w / bin_width
        avail_height = max_h
        packer.add_bin(bin_width, avail_height)

        packer.pack()

        packer[0]

        all_rects = packer.rect_list()
        all_rects = np.array(all_rects)

        rids = all_rects[:, 5]
        tl_x = all_rects[:, 1]
        tl_y = all_rects[:, 2]
        w = all_rects[:, 3]
        h = all_rects[:, 4]

        ltrb = kwimage.Boxes(all_rects[:, 1:5], 'xywh').to_ltrb()
        canvas_w, canvas_h = ltrb.data[:, 2:4].max(axis=0)

        canvas = np.zeros((canvas_h, canvas_w), dtype=np.float32)

        for b, x, y, w, h, rid in all_rects:
            img = images[rid]
            img = kwimage.ensure_float01(img)
            canvas, img = kwimage.make_channels_comparable(canvas, img)
            canvas[y: y + h, x: x + w] = img

        kwplot.imshow(canvas)

    """
